#ifndef _GRN_DATATYPE_H_
#define _GRN_DATATYPE_H_

#include <string>

#include "Pointer.h"
#include "Visitor.h"

#define DECLARE_DATATYPE(X) public:																	\
							static const DataType Type;												\
							typedef tFactory<X> Factory; 											\
							virtual const DataType& GetDataType() const { return Type; }			\
							virtual void Accept( BaseVisitor* aV ) { AcceptImpl( this, aV); }
							
#define IMPLEMENT_DATATYPE(X) const DataType grn::X::Type(#X);

namespace grn
{
	
	typedef std::string DataType;
	
	class DataObject: public RefCounter, public BaseVisitable
	{
	public:
		
		class Factory
		{
		public: 
			virtual ~Factory() {}
			virtual DataObject* New( ) const = 0;
			virtual const std::string& getName() const = 0;
		};

		template<class M> class tFactory : public Factory
		{
		public: 
			virtual M* New() const { return new M(); }
			virtual const std::string& getName() const { return M::Type; }
		};
		
		
	public:
		virtual const DataType& GetDataType() const = 0;
		virtual void Accept( BaseVisitor* aV ) = 0;

	};

	template< class C > 
	Pointer<C> down_cast( DataObject* aSource ) { return static_cast<C*> (aSource); }
}

#endif /*_DATATYPE_H_*/
