#ifndef _GRN_PIPELINE_H_
#define _GRN_PIPELINE_H_

#include <iostream>
#include <vector>
#include <map>
#include <list>

#include "DataType.h"
#include "Module.h"
#include "engine/Engine.h"
#include "ModuleDraw.h"


#define  GRN_RegisterDataType(X) 	GetDataTypeID( X::Type ); \
									iObjectFactories.push_back( new X::Factory() );
	


namespace grn
{
	
	
	class Pipeline : public Module
	{
		DECLARE_MODULE(Pipeline);
	
	public:

		Pipeline( Pipeline* aParent );
		virtual ~Pipeline() {}
	
		void Draw(ModuleDrawVisitor& lDrawer);
		
		void SetSelected( Module* aModule );
		void RemoveSelected();

		/** Return the code for a given data type. If not present, add it and return the new code. */
		int GetDataTypeID( DataType aType );
		
		/** Return the data type for a given code. */
		DataType FindDataType( int );
		
		typedef std::map< DataType, int > TypeMap;
		TypeMap& GetTypeMap();

		/** Find the factory for a given type of module. */
		Module::Factory* FindFactory( std::string aModuleType );
		
		/** Find the factory for a given type of module. */
		DataObject::Factory* FindObjectFactory( std::string aDataType );
		
		virtual void Link( Engine* aEngine );
		virtual void Setup();
		virtual void Run();
		
		void Emit( Module::OutputPad* aPad, DataObject* aData );
		void EmitEOD( Module::OutputPad* aPad );
		
		void Clear();
		void Load( const std::string& aFileName );
		void Save( const std::string& aFileName ) const;

		/** Redefined from Module to clear owned modules too. */
		void ClearData();

		void RemoveLinksTo( Module::InputPad* aPad );

		/** Output serializer.*/
		virtual void Serialize( std::ostream& s ) const;

		/** Input serializer.*/
		virtual void Deserialize( std::istream& s );

		std::string GetName() const { return iName; }
		void SetName( std::string aName ) { iName = aName; }
		
		/** Get a user readable name. */
		virtual std::string GetModuleUserName() { return "Pipeline: "+iName; }

		/** Find a pipeline by name. */
		virtual Pipeline* FindPipeline( std::string aName );

		/** Enable or diable this pipeline. */
		void SetEnabled( bool aEnabled ) { iEnabled = aEnabled; }
		bool GetEnabled( ) { return iEnabled; }

		void LazySetEnabled( bool aEnabled ) { iLazyToggleEnabled = (iEnabled!=aEnabled); }
		void UpdateEnabled() { if (iLazyToggleEnabled) iEnabled = !iEnabled; iLazyToggleEnabled=false;}

		/** Set start status: enabled or diabled. */
		void SetStartEnabled( bool aEnabled ) { iStartEnabled = aEnabled; }
		bool GetStartEnabled( ) { return iStartEnabled; }

	public:
		typedef std::vector<Module::Factory*> ModuleFactoryList;
		ModuleFactoryList iModuleFactories;

		typedef std::vector<DataObject::Factory*> ObjectFactoryList;
		ObjectFactoryList iObjectFactories;

		typedef std::vector<Module*> ModuleList;
		ModuleList iModules;
		Module* iSelected;
		
		std::list<Module*> iRunQueue;
		
	protected:

		/** User name for this pipeline. */
		std::string iName;
		
		/** Currently enabled? */
		bool iEnabled;
		bool iLazyToggleEnabled;
		
		/** Initially enabled? */
		bool iStartEnabled;
		
		int iNextDataTypeID;
		TypeMap iTypeMap;
		
			
	};


}

#endif /*_PIPELINE_H_*/
