#ifndef _GRN_BASICTYPES_H_
#define _GRN_BASICTYPES_H_


#include "../DataType.h"

#include "AABB.h"
#include "SimdTransform.h"
#include "SimdVector3.h"
#include "SimdVector4.h"

#include <list>

namespace grn
{
	
	class Boolean: public DataObject
	{
		DECLARE_DATATYPE(Boolean)

	public:		
		/** Err.... value of this integer. */
		bool iValue;
		
	};


	class Integer: public DataObject
	{
		DECLARE_DATATYPE(Integer)
		
	public:
		/** Err.... value of this integer. */
		int iValue;
		
	};


	class Real: public DataObject
	{
		DECLARE_DATATYPE(Real)

	public:		
		/** Err.... value of this real number. */
		float iValue;
		
	};


	class String: public DataObject
	{
		DECLARE_DATATYPE(String)

	public:
		/** Err.... value of this string. */
		std::string iValue;
		
	};


	class StringList: public DataObject
	{
		DECLARE_DATATYPE(StringList)

	public:
		/** Err.... value of this string. */
		std::vector<std::string> iValue;
		
	};


	class Vector: public DataObject
	{
		DECLARE_DATATYPE(Vector)

	public:
		
		SimdVector3 iValue;
		
	};


	class Color: public DataObject
	{
		DECLARE_DATATYPE(Color)

	public:
		
		SimdVector4 iColor;
		
	};


	class BoundingVolume: public DataObject
	{
		DECLARE_DATATYPE(BoundingVolume)

	public:
		
		AABB iAABB;
		
	};


	class Transform: public DataObject
	{
		DECLARE_DATATYPE(Transform)

	public:
		
		/** Rotate-translate decomposed transform. First rotate, then translate.*/
		SimdTransform iTransform;
		
		Transform();
		SimdTransform GetTransform() { return iTransform; }

		void SetIdentity();

	};


	class Group: public DataObject
	{
		DECLARE_DATATYPE(Group)

	public:

		class Collection
		{
		public:
			Collection( const std::string& aN ) : iName(aN) {}
			
			std::string iName;
			std::vector< Pointer<DataObject> > iObjects;
		};
				
		std::vector< Collection > iCollections;
		
	};

}


#endif /*_BASICTYPES_H_*/
