#ifndef _GRN_MATERIAL_H_
#define _GRN_MATERIAL_H_


#include "DataType.h"

#include "Camera.h"
#include "BasicTypes.h"
#include "RenderTarget.h"

#include <vector>


namespace grn
{
	
	class Material: public DataObject
	{
		DECLARE_DATATYPE(Material);

	public:
		virtual ~Material() {}
		
		// Pipeline parameters that can be set when the material is not activated
		virtual void SetTextureParameter( std::string, Texture* ) {};
		virtual void SetCameraParameter( std::string, Camera* ) {};

	};
	
	
	class DepthCodeMaterial : public Material
	{
		DECLARE_DATATYPE(DepthCodeMaterial);
	};
	
	
	class UVCodeMaterial : public Material
	{
		DECLARE_DATATYPE(UVCodeMaterial);
	};
	
	
	class PhongMaterial : public Material
	{
		DECLARE_DATATYPE(PhongMaterial);

	public:
		PhongMaterial( bool aReceiveShadows=true );
		
		virtual void SetTextureParameter( std::string aParamName, Texture* aTexture );
		virtual void SetCameraParameter( std::string aParamName, Camera* aCam );
		
		float iExponent;
		Color iSpecular;
		Color iDiffuse;
		Color iAmbient;
		bool iReceiveShadows;

		// run-time info		
		Pointer<Texture> rDepthBuffer;
		SimdMatrix4x4 rDepthMatrix;
		SimdVector3 iEyePosition;
		SimdVector3 iLightPosition;

	};
	
}


#endif /*_MESH_H_*/
