#ifndef _GRN_MESH_H_
#define _GRN_MESH_H_


#include "../DataType.h"

#include "AABB.h"

#include <vector>

#define GRN_MAX_TEX_COORDS_PER_VERTEX	8

namespace grn
{
	class Mesh: public DataObject
	{
		DECLARE_DATATYPE(Mesh);

	public:
		class VertexBuffer
		{
		public:
			/** Buffer handle if uploaded to video memory. 0 otherwise. */
			unsigned int iHandle;

			/** Pointer to the raw data. */
			char* iData;
			
			/** Size if bytes of the buffer. */
			int iSize;

			/** Semantics of the data. Used to link with shader data. */
			enum Semantic { ePosition=10, eNormal, eTexCoord0, eTexCoord1, eTexCoord2, eTexCoord3, eTexCoord4 } iSemantic;
			
			/** Data type. */
			enum DataType { eFloat = 0, eByte, _eNumTypes } iDataType;
			
			/** Memory sizes for each data type. */
			static const int sTypeSizes[_eNumTypes];
			
			/** Number of elements of iType for each vertex. */
			unsigned char iNumElements;
		};
		
		class IndexBuffer
		{
		public:
			/** Buffer handle if uploaded to video memory. 0 otherwise. */
			unsigned int iHandle;

			/** Pointer to the raw data. */
			char* iData;
			
			/** Size if bytes of the buffer. */
			int iSize;
			
			/** Data type. */
			enum DataType { eInt = 0, eShort, eByte, _eNumTypes } iDataType;
			
			/** Memory sizes for each data type. */
			static const int sTypeSizes[_eNumTypes];
		};
		
	public:
		Mesh();
		virtual ~Mesh();
		
		int iNumFaces;
		int iNumVertices;
		std::vector<VertexBuffer*> iVertexBuffers;
		IndexBuffer* iIndexBuffer;
		AABB iAABB;

	};
}


#endif /*_MESH_H_*/
