

#include "Object.h"
#include "BasicTypes.h"

#include <algorithm>

using namespace grn;


const DataType Object::Type("Girona::Object");

const std::string kNullObjectName("<null>");


Object::Object()
{
	iPhysicsController = NULL;
}


Object::~Object()
{
}


void Object::UpdateBoundingVolume()
{
	//@TODO: Move this code to the bounding box class
	AABB lNew;
	
	SimdPoint3 lCorners[8];
	iMesh->iAABB.GetCorners(lCorners);
	for ( int i=0; i<8; i++ )
		lNew.Bound( iInstanceTransform->GetTransform()(lCorners[i] ) );
	iAABB = lNew;
}


void Object::GetWorldTransform( SimdTransform& aTransform )
{
	
	if (iParent) iParent->GetWorldTransform( aTransform );
	else aTransform.setIdentity();

    //aTransform.iTranslation = SimdMatrix3x3(aTransform.iRotation) * iInstanceTransform->iTranslation + aTransform.iTranslation;
 	//aTransform.iRotation = aTransform.iRotation * iInstanceTransform->iRotation;
 	aTransform = aTransform * iInstanceTransform->iTransform;
 
}


void Object::SetWorldTransform( const SimdTransform& aWorldTransform )
{
	SimdTransform lWorldTransform = aWorldTransform;
	
	if (iParent)
	{
		SimdTransform lParentTransform;
		iParent->GetWorldTransform(lParentTransform);
		
		SimdTransform lParentTransformInverted = lParentTransform.inverse();
		
		lWorldTransform = lParentTransformInverted * aWorldTransform;
	}
	
	iInstanceTransform->iTransform = lWorldTransform;
	
	//NotifyAll();
}


void Object::SetWorldOrientation( float euler1, float euler2, float euler3 )
{
	SimdTransform lWorldTransform;
	GetWorldTransform( lWorldTransform );
	lWorldTransform.setRotation( SimdQuaternion(euler1, euler2, euler3) );
	
	if (iParent)
	{
		SimdTransform lParentTransform;
		iParent->GetWorldTransform(lParentTransform);
		
		SimdTransform lParentTransformInverted = lParentTransform.inverse();
		
		lWorldTransform = lParentTransformInverted * lWorldTransform;
	}
	
	iInstanceTransform->iTransform = lWorldTransform;
	
	NotifyAll();
	
}


void Object::Relink( Pointer<Object> aParent )
{
	
	SimdTransform lWorldTransform;
	GetWorldTransform(lWorldTransform);
	
	if (iParent) iParent->iChildren.erase(std::remove(iParent->iChildren.begin(), iParent->iChildren.end(), this), iParent->iChildren.end());

	iParent = aParent;

	if (iParent) iParent->iChildren.push_back(this);

	SetWorldTransform(lWorldTransform);
	
}


void Object::Updated(Object* rParent)
{
	if (rParent==iParent)
	{
		// Update the physics representation
		if (iPhysicsController) 
		{
			SimdTransform lTrans;
			GetWorldTransform(lTrans);
			SimdQuaternion lRot = lTrans.getRotation();
			SimdPoint3 lPos = lTrans.getOrigin();
			
			iPhysicsController->setOrientation( lRot.x(), lRot.y(), lRot.z(), lRot[3] );
			iPhysicsController->setPosition( lPos.x(), lPos.y(), lPos.z() );
		}
		NotifyAll();
	}
}

void Object::NotifyAll()
{
	for ( std::list<ObjectObserver*>::iterator it = iChildren.begin()
		; it != iChildren.end()
		; it++ )
		(*it)->Updated(this);
}


bool Object::HasProperty( const std::string& aProperty ) const
{
	if (iProperties)
		return ( std::find(iProperties->iValue.begin(), iProperties->iValue.end(), aProperty) != iProperties->iValue.end() ) ;
	else return false;
}

