#ifndef _GRN_OBJECT_H_
#define _GRN_OBJECT_H_


#include "../DataType.h"

#include "BasicTypes.h"
#include "Mesh.h"
#include "Material.h"
#include "AABB.h"

#include "CcdPhysicsController.h"

#include <list>
#include <vector>

namespace grn
{
	static const std::string kNullObjectName;

	class Object;
	
	class ObjectObserver
	{
	public:
		virtual ~ObjectObserver(){};
		virtual void Updated(Object* rParent)=0;
	};


	class Object: public DataObject, public ObjectObserver
	{
		DECLARE_DATATYPE(Object);

	public:
		Object();
		virtual ~Object();
		
		Pointer<String> iName;
		Pointer<Mesh> iMesh;
		Pointer<Material> iMaterial;

		/** Hierarchy information. The parent nameis only for delayerd hierarchy re-link. */
		Pointer<Object> iParent;
		Pointer<String> iParentName;
		Pointer<StringList> iProperties;

		/** Transformation obeservers. Non owned pointers. Usually children. */
		std::list<ObjectObserver*> iChildren;

		CcdPhysicsController* iPhysicsController;

		/** Bounding volume of the object after applying the instance transform. */
		AABB iAABB;

		Pointer<Transform> iInstanceTransform;

		/** Update bounding box. */
		void UpdateBoundingVolume();
		
		void GetWorldTransform( SimdTransform& aTransform );
		void SetWorldTransform( const SimdTransform& aTransform );
		void SetWorldOrientation( float euler1, float euler2, float euler3 );
		
		/** Change or set parent. Keeps the exact world transform. */
		void Relink( Pointer<Object> aParent );

		/** Find if the object has the given property. */
		bool HasProperty( const std::string& aProperty ) const;

	protected:
		/** Update all observers.*/
		void NotifyAll();
	
	// ObjectOserver redefinitions		
	public:
		virtual void Updated(Object* rParent);

	};
	
}


#endif /*_MESH_H_*/
