#ifndef _GRN_PHYSICS_ENVIRONMENT_H_
#define _GRN_PHYSICS_ENVIRONMENT_H_


#include "DataType.h"

#include "Object.h"
#include "CcdPhysicsEnvironment.h"
#include "SimdVector3.h"

#include <map>
#include <vector>

namespace grn
{
	
	class PhysicsProperties: public DataObject
	{
		DECLARE_DATATYPE(PhysicsProperties);
		
	public:

		bool iActor;
		bool iDynamic;
		bool iRigidBody;
		bool iGhost;
		bool iDeformable;

		float iMass;
		float iFriction;
		float iRestitution;
		float iInertia;
		float iRadius;
		float iLinearDrag;
		float iAngularDrag;
		bool iHasAnisotropicFriction;
		SimdVector3 iAnisotropicFrictionScaling;
		bool iDoFH;
		bool iDoRotFH;
		
		SimdVector3 iLocalBoundsCenter;
		SimdVector3 iLocalBoundsExtends;
		
		typedef enum { eBoundNone, eBoundBox, eBoundPolyt, eBoundPolyh, eBoundSphere, eBoundCylinder, eBoundCone } BoundType;
		BoundType iBoundType;
		bool iConcave;
				
	};
		
	class PhysicsEnvironment: public DataObject
	{
		DECLARE_DATATYPE(PhysicsEnvironment);

	public:
 
		PhysicsEnvironment();
		virtual ~PhysicsEnvironment();
		
		virtual void Update( float aCurrentTime, float aDeltaTime );
		
		/** Physics engine. */
		CcdPhysicsEnvironment* iPhysicsEnvironment;

		/** This raw data is stored to allow serialization. */
		std::vector< std::pair< Pointer<Object>, Pointer<PhysicsProperties> > > iRawData;
		
		void Build();
	};
	
	
	class CollisionList: public DataObject
	{
		DECLARE_DATATYPE(CollisionList);
		
	public:
	
		typedef std::pair<const void*, const void*> CollisionPair;
		
		std::vector< CollisionPair > iPairs;
	};


	class ScenePhysicsProperties: public DataObject
	{
		DECLARE_DATATYPE(ScenePhysicsProperties);
		
	public:
		SimdVector3 iGravity;
		
	};


}


#endif
