#ifndef _GRN_RENDERTARGET_H_
#define _GRN_RENDERTARGET_H_


#include "DataType.h"
#include "engine/Engine.h"


class RenderTexture;

namespace grn
{
	

	class RenderTarget: public DataObject
	{
		DECLARE_DATATYPE(RenderTarget);

	public:
		
		// This methods should be pure virtual, but this is incompatible with the macro above.
		
		virtual void Activate() { assert(false); };
		virtual void Deactivate() { assert(false); };
		virtual void Present() { assert(false); };
		
		virtual void Clear() { assert(false); }
		
	}; 

	
	class ScreenRenderTarget : public RenderTarget
	{
		DECLARE_DATATYPE(ScreenRenderTarget);
	public:
		ScreenRenderTarget() { iValid = false; }
		ScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine );

		bool iValid;			
		int iX,iY,iWidth,iHeight;

		virtual void Present();
		
	protected:
		Engine* rEngine;
	};


	class Texture: public DataObject
	{
		DECLARE_DATATYPE(Texture);

	public:
		
		// This methods should be pure virtual, but this is incompatible with the macro above.
		virtual void Bind() { assert(false); }
		virtual void Unbind() { assert(false); }
		
	}; 

	
	class OffscreenRenderTarget : public RenderTarget
	{
		DECLARE_DATATYPE(OffscreenRenderTarget);

	public:
		
		// Has to be created by the renderer
		OffscreenRenderTarget() { iValid = false; }
		OffscreenRenderTarget( int aWidth, int aHeight );

		bool iValid;
		int iWidth,iHeight;
		
		virtual Texture* GetColorTexture() { return NULL; }
		virtual Texture* GetDepthTexture() { return NULL; }
		
	};

}


#endif /*_BASICTYPES_H_*/
