#ifndef _GRN_ENGINE_H_
#define _GRN_ENGINE_H_


#include <vector>
#include <string>
#include <map>

namespace grn
{
	class RootPipeline;
	class Renderer;
	
	class Engine
	{
	public:
		typedef enum { eSpace=0, eLeft, eRight, eUp, eDown, eEscape, eQ, eW, eE, eA, eS, eD, eZ, eX, eC, _eNumKeys } KeyCode;
		static const char* sKeyNames[_eNumKeys];
	
	public:
	
		class MouseListener
		{
		public:
			virtual ~MouseListener() {}
			virtual void ProcessMouse( int aDeltaX, int aDeltaY ) = 0;
		};
	
		class KeyListener
		{
		public:
			virtual ~KeyListener() {}
			virtual void ProcessKey( KeyCode aKey, bool aPressed ) = 0;
		};
	
	public:

		Engine( RootPipeline* aP, Renderer* aR ) : rRootPipeline(aP), rRenderer(aR) {}
		virtual ~Engine() {}

		virtual void Start();
		virtual void Stop();
		
		virtual void RunFrame();
		
		virtual void Build( const std::string& aFile );
		virtual void LoadData( const std::string& aFile );

		virtual void SwapBuffers()=0;
			
		virtual void Register( grn::Engine::MouseListener* aListener )
		{
			iMouseListeners.push_back( aListener );
		}

		virtual void Register( unsigned short aKey, grn::Engine::KeyListener* aListener )
		{
			iKeyListeners.insert( std::pair<unsigned short,KeyListener*>( aKey,aListener) );
		}

		virtual void SetScreenArea(int aX, int aY, int aWidth, int aHeight)
		{
			iX = aX; iY = aY; iWidth = aWidth; iHeight = aHeight;
		}
	
		virtual void GetScreenArea(int& aX, int& aY, int& aWidth, int& aHeight)
		{
			aX = iX; aY = iY; aWidth = iWidth; aHeight = iHeight;
		}
		
		virtual float GetDeltaTime() { return iDeltaTime; }
		virtual void SetDeltaTime( float aDeltaTime ) { iDeltaTime = aDeltaTime; }
	
		virtual float GetCurrentTime() { return iCurrentTime; }
		virtual void SetCurrentTime( float aCurrentTime ) { iCurrentTime = aCurrentTime; }
	
		virtual void ProcessKey( KeyCode aCode, bool aPressed );
	
		inline Renderer* GetRenderer() { return rRenderer; }
	
	protected:
		std::vector< MouseListener* > iMouseListeners;
		typedef std::multimap< unsigned short, KeyListener* > KeyListeners;
		KeyListeners iKeyListeners;
	
		/** Screen rendering area. */
		int iX,iY,iWidth, iHeight;
		
		/** Time in seconds since last frame. */
		float iDeltaTime;
		float iCurrentTime;
		
		RootPipeline* rRootPipeline;
		
		/** Reference to the rendering device. */
		Renderer* rRenderer;
				
	};

}

#endif /*_MODULE_H_*/
