#ifndef PHONG_SHADERS_H_
#define PHONG_SHADERS_H_



const char* PIXEL_PHONG_V_SHADER = 
"\
attribute vec3 VertexNormal;\
\
uniform vec3 uEyePos;\
uniform vec3 uLightPos;\
uniform vec4 uAmbient;\
uniform vec4 uDiffuse;\
uniform vec4 uSpecular;\
uniform float uExponent;\
uniform mat4 uModelTransform;\
uniform mat4 uModelNormalTransform;\
\
varying vec3 oWorldPosition;\
varying vec4 oClipPosition;\
varying vec3 oNormal;\
varying vec3 oLightVector; \
\
void main(void)\
{\
	oNormal = uModelNormalTransform * vec4(VertexNormal,1.0);\
	oWorldPosition = (uModelTransform * gl_Vertex).xyz;\
	oLightVector = uLightPos - oWorldPosition; \
\
	gl_Position = gl_ModelViewProjectionMatrix * vec4(oWorldPosition, 1.0);\
	oClipPosition = gl_Position;\
}\
";


const char* SHADOW_PIXEL_PHONG_V_SHADER = 
"\
attribute vec3 VertexNormal;\
\
uniform vec3 uEyePos;\
uniform vec3 uLightPos;\
uniform vec4 uAmbient;\
uniform vec4 uDiffuse;\
uniform vec4 uSpecular;\
uniform float uExponent;\
uniform mat4 uModelTransform;\
uniform mat4 uModelNormalTransform;\
uniform mat4 uDepthTransform;\
\
varying vec3 oWorldPosition;\
varying vec4 oClipPosition;\
varying vec4 oShadowPosition;\
varying vec3 oNormal;\
varying vec3 oLightVector; \
\
void main(void)\
{\
	oNormal = uModelNormalTransform * vec4(VertexNormal,1.0);\
	oWorldPosition = (uModelTransform * gl_Vertex).xyz;\
	oLightVector = uLightPos - oWorldPosition; \
\
	gl_Position = gl_ModelViewProjectionMatrix * vec4(oWorldPosition, 1.0);\
	oClipPosition = gl_Position;\
	oShadowPosition = uDepthTransform * vec4(oWorldPosition, 1.0);\
}\
";


const char* PHONG_F_COMMON = 
" \
\
uniform vec3 uEyePos;\
uniform vec3 uLightPos;\
uniform vec4 uAmbient;\
uniform vec4 uDiffuse;\
uniform vec4 uSpecular;\
uniform float uExponent;\
uniform sampler2D uDepthBuffer;\
\
varying vec3 oNormal; \
varying vec3 oWorldPosition; \
varying vec4 oClipPosition; \
varying vec3 oLightVector; \
varying vec4 oShadowPosition; \
\
float4 phong( vec3 aNormal, vec3 aLightVector, vec3 aWorldPosition )\
{\
	vec3 N = normalize( aNormal );\
	vec3 L = normalize( aLightVector );\
	vec3 E = normalize( uEyePos - aWorldPosition );\
	vec3 H = (L+E) * 0.5;\
\
	float diffuseFactor = clamp( dot(N,L), 0.0, 1.0 );\
	float specFactor = uExponent * pow(clamp( dot(N,H), 0.0, 1.0 ), uExponent);\
\
	float4 Color = uAmbient \
				 + uDiffuse * diffuseFactor \
			 	 + uSpecular * specFactor; \
	return Color;\
}\
\
\
float4 shadow( float4 aShadowPosition, float4 aColor )\
{\
	if (abs(aShadowPosition.s) < abs(aShadowPosition.q) && \
		abs(aShadowPosition.t) < abs(aShadowPosition.q) && \
		aShadowPosition.p > 0.0) \
	{ \
		float Depth = aShadowPosition.z/aShadowPosition.w*0.5+0.5;\
		float4 MapDepthV = texture2D(uDepthBuffer, (aShadowPosition.xy/aShadowPosition.w * 0.5 + 0.5));\
		float MapDepth = MapDepthV.x + MapDepthV.y/256.0; \
		\
		if (Depth>MapDepth+0.004) aColor = uAmbient;\
	}\
	return aColor;\
}\
\
\
";


const char* PIXEL_PHONG_F_SHADER_BODY = 
" \
void main(void)\
{\
	gl_FragColor = phong( oNormal, oLightVector, oWorldPosition ); \
}\
";


const char* SHADOW_PIXEL_PHONG_F_SHADER_BODY = 
" \
void main(void)\
{\
	gl_FragColor = shadow( oShadowPosition, phong( oNormal, oLightVector, oWorldPosition ) ); \
}\
";


#define PIXEL_PHONG_F_SHADER std::string(PHONG_F_COMMON) + std::string(PIXEL_PHONG_F_SHADER_BODY)
#define SHADOW_PIXEL_PHONG_F_SHADER std::string(PHONG_F_COMMON) + std::string(SHADOW_PIXEL_PHONG_F_SHADER_BODY)


#endif /*PHONG_SHADERS_H_*/
