
#include <cassert>
#include <cstdio>
#include <iostream>


#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glext.h>


#include "Program.h"
#include "datatypes/Mesh.h"



using namespace grn;


Program::Program()
: iVertexShaderHandle(0)
, iFragmentShaderHandle(0)
, iProgramHandle(0)
{
}

Program::~Program()
{
}

GLcharARB* ReadFile( std::string aFile )
{
	FILE* lF = fopen( aFile.c_str(), "rb" );
	
	if (lF==NULL) return NULL;
	
	fseek( lF, 0, SEEK_END );
	long lSize = ftell(lF);
	char* lRes = (GLcharARB*) malloc( lSize+1 );
	fseek( lF, 0, SEEK_SET );
	
	fread( lRes, lSize, 1, lF );
	lRes[lSize] = 0;
	
	fclose(lF);
	
	return lRes;
}


GLhandleARB Program::BuildShaderFromString( std::string aShader, GLenum aType )
{
	GLhandleARB lShaderHandle = glCreateShaderObjectARB( aType );
	assert(lShaderHandle!=0);
	
	const GLcharARB* lCode = aShader.c_str();
	
	glShaderSourceARB( lShaderHandle, 1, (const GLcharARB**)&lCode, NULL );
		
	glCompileShaderARB( lShaderHandle );
	
	GLint lStatus;
	glGetObjectParameterivARB( lShaderHandle, GL_OBJECT_COMPILE_STATUS_ARB, &lStatus );
	if (lStatus==1) std::cout << " [GRN] Program compiled ok."<<std::endl;
	else 
	{
		std::cout<<" [GRN] Program  compilation failed. "<<std::endl;
		GLcharARB lMessage[0xFFF];
		glGetInfoLogARB( lShaderHandle, 0xFFF, NULL, lMessage );
		std::cout<<lMessage<<std::endl;
	}
	
	return lShaderHandle;
}


GLhandleARB Program::BuildShader( std::string aFile, GLenum aType )
{
	GLcharARB* lCode = ReadFile( aFile );
	
	GLhandleARB lShaderHandle = BuildShaderFromString( lCode, aType );
	
	free( lCode );
	
	return lShaderHandle;
}


void Program::LoadFromStrings( std::string aVertexShader, std::string aFragmentShader )
{
	iVertexShaderHandle = BuildShaderFromString( aVertexShader, GL_VERTEX_SHADER_ARB );
	CHECK_GL_ERROR(" [GRN] After vertex compile : ");	    
	iFragmentShaderHandle = BuildShaderFromString( aFragmentShader, GL_FRAGMENT_SHADER_ARB );	
	CHECK_GL_ERROR(" [GRN] After pixel compile : ");	    
	Link();
	CHECK_GL_ERROR(" [GRN] After link : ");	    
}


void Program::LoadFromFiles( std::string aVertexFile, std::string aFragmentFile )
{
	iVertexShaderHandle = BuildShader( aVertexFile, GL_VERTEX_SHADER_ARB );
	iFragmentShaderHandle = BuildShader( aFragmentFile, GL_FRAGMENT_SHADER_ARB );	
	Link();
}


void Program::Link(  )
{
	CHECK_GL_ERROR(" [GRN] link 0 : ");	    
	iProgramHandle = glCreateProgramObjectARB();
	assert(iProgramHandle);
	
	CHECK_GL_ERROR(" [GRN] link 1 : ");	    

	//glBindAttribLocationARB( iProgramHandle, 16, "VertexPosition" );
	glBindAttribLocationARB( iProgramHandle, Mesh::VertexBuffer::eNormal, "VertexNormal" );
	glBindAttribLocationARB( iProgramHandle, Mesh::VertexBuffer::eTexCoord0, "VertexTex0" );
	//glBindAttribLocationARB( iProgramHandle, 18, "VertexColor" );
	//glBindAttribLocationARB( iProgramHandle, 19, "VertexTexCoord" );
	
	SKIP_GL_ERROR;

	glAttachObjectARB( iProgramHandle, iVertexShaderHandle );
	glAttachObjectARB( iProgramHandle, iFragmentShaderHandle );
	glLinkProgramARB( iProgramHandle );
	
	CHECK_GL_ERROR(" [GRN] link 3 : ");	    
	
	GLint lStatus;
	glGetObjectParameterivARB( iProgramHandle, GL_OBJECT_LINK_STATUS_ARB, &lStatus );	
	if (lStatus==1) std::cout << " [GRN] Program linked ok."<<std::endl;
	else 
	{
		std::cout<<" [GRN] Program linking failed."<<std::endl;
		GLcharARB lMessage[0xFFF];
		glGetInfoLogARB( iProgramHandle, 0xFFF, NULL, lMessage );
		std::cout<<lMessage<<std::endl;
	}
	
	glValidateProgramARB( iProgramHandle );
	
	glGetObjectParameterivARB( iProgramHandle, GL_OBJECT_VALIDATE_STATUS_ARB, &lStatus );	
	if (lStatus==1) std::cout << " [GRN] Program validated ok."<<std::endl;
	else 
	{
		std::cout<<" [GRN] Program validation failed."<<std::endl;
		GLcharARB lMessage[0xFFF];
		glGetInfoLogARB( iProgramHandle, 0xFFF, NULL, lMessage );
		std::cout<<lMessage<<std::endl;
	}

}


void Program::Activate()
{
	CHECK_GL_ERROR(" [GRN] Before activating : ");	    
	glUseProgramObjectARB( iProgramHandle );
	CHECK_GL_ERROR(" [GRN] Activating : ");	    
}

void Program::Deactivate()
{
	glUseProgramObjectARB( 0 );
}


Program::Parameter Program::FindParameter( const std::string& aName ) const
{
	return glGetUniformLocationARB( iProgramHandle, aName.c_str() );
}


void Program::SetParameterValue( const Parameter& aParam, const SimdMatrix4x4& aValue)
{
	float temp[16];
	aValue.getValue(temp);
	
	glUniformMatrix4fvARB( aParam, 1, false, temp );
}


void Program::SetParameterValue( const Parameter& aParam, const SimdVector3& aValue)
{
	glUniform3fARB( aParam, aValue.x(), aValue.y(), aValue.z() );
}


void Program::SetParameterValue( const Parameter& aParam, const SimdVector4& aValue)
{
	glUniform4fARB( aParam, aValue.x(), aValue.y(), aValue.z(), aValue.w() );
}


void Program::SetParameterValue( const Parameter& aParam, const GLfloat aValue)
{
	glUniform1fARB( aParam, aValue );
}


void Program::SetParameterValue( const Parameter& aParam, const GLint aValue)
{
	glUniform1iARB( aParam, aValue );
	CHECK_GL_ERROR(" [GRN] Setting program parameter : ");	    
	
}



