#ifndef _GRN_PROGRAM_H_
#define _GRN_PROGRAM_H_


#include <string>


#ifdef WIN32
#include <windows.h>
#endif // WIN32
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "SimdMatrix4x4.h"
#include "SimdVector3.h"
#include "SimdVector4.h"


#define CHECK_GL_ERROR(M) { GLint iErr = glGetError(); if (iErr!=GL_NO_ERROR) std::cout<<M<<" : "<<gluErrorString(iErr)<<std::endl; }
#define SKIP_GL_ERROR { glGetError(); }


namespace grn
{

#ifndef GIRONA_PSP

	class Program
	{
	public:
	
		Program();
		virtual ~Program();
		
		void LoadFromFiles( std::string aVertexFile, std::string aFragmentFile );
		void LoadFromStrings( std::string aVertexFile, std::string aFragmentFile );
		
		void Activate();
		void Deactivate();
		
		typedef GLint Parameter;
		static const Parameter kInvalidParameter = -1;
		
		Parameter FindParameter( const std::string &aName ) const;
		
		void SetParameterValue( const Parameter& aParam, const SimdMatrix4x4& aValue);
		void SetParameterValue( const Parameter& aParam, const SimdVector4& aValue);
		void SetParameterValue( const Parameter& aParam, const SimdVector3& aValue);
		void SetParameterValue( const Parameter& aParam, const GLfloat aValue);
		void SetParameterValue( const Parameter& aParam, const GLint aValue);
		
	protected:
	
		GLhandleARB iVertexShaderHandle;
		GLhandleARB iFragmentShaderHandle;
		GLhandleARB iProgramHandle;

		GLhandleARB BuildShader( std::string aFile, GLenum aType ); 
		GLhandleARB BuildShaderFromString( std::string aFile, GLenum aType ); 
		void Link();
	
	};

#else

	class Program
	{
	public:
	
		Program() {}
		virtual ~Program() {}
		
		inline void LoadFromFiles( std::string aVertexFile, std::string aFragmentFile ) {}
		inline void LoadFromStrings( std::string aVertexFile, std::string aFragmentFile ) {}
		
		inline void Activate() {}
		inline void Deactivate() {}
		
		typedef GLint Parameter;
		static const Parameter kInvalidParameter = -1;
		
		inline Parameter FindParameter( const std::string &aName ) const { return 0; }
		
		inline void SetParameterValue( const Parameter& aParam, const SimdTransform& aValue) {}
		inline void SetParameterValue( const Parameter& aParam, const SimdVector4& aValue) {}
		inline void SetParameterValue( const Parameter& aParam, const SimdVector3& aValue) {}
		inline void SetParameterValue( const Parameter& aParam, const GLfloat aValue) {}
		inline void SetParameterValue( const Parameter& aParam, const GLint aValue) {}
	
	};

	// dummy implementation of a missing function in PSP-GL
	inline const GLubyte* gluErrorString(GLenum) { return (GLubyte*)("gluErrorString not implemented."); }


#endif


};

#endif //_PROGRAM_H_

