#ifndef GRN_RENDERER_GL2_H_
#define GRN_RENDERER_GL2_H_


#include "engine/Renderer.h"

#include "Program.h"


namespace grn
{
	
	class GL2OffscreenRenderTarget : public OffscreenRenderTarget
	{
	public:
		GL2OffscreenRenderTarget( int aWidth, int aHeight, bool aColor=true, bool aDepth=true );
			
		virtual void Activate();
		virtual void Deactivate();
		virtual void Present();
		virtual void Clear();
		
		virtual Texture* GetColorTexture();
		virtual Texture* GetDepthTexture();

		GLuint iFrameBuffer, iDepthRenderBuffer, iTex;
		
	};
	
	
	class GL2ScreenRenderTarget : public ScreenRenderTarget
	{
	public:
		GL2ScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine );

		virtual void Activate();
		virtual void Deactivate();
		
		virtual void Clear();
	};
	
	
	class GL2Texture : public Texture
	{
	public:
		virtual void Bind();
		virtual void Unbind();
		
		GLenum iTextureTarget;
		GLint iTextureID;
	};
	

	class RendererGL2
		: public Renderer
	{
	public:
		RendererGL2();
		
		void Activate( const Camera& aCam ) const;
		void Activate( Material* aMat );
		void Deactivate( Material* aMat );
		void SetModelTransform( const SimdTransform&, const SimdTransform& );

		void Render( const AABB& aB ) const;
		void RenderInverted( const AABB& aB ) const;
		
		/** Upload all buffers to the video memory. */
		void Upload( Mesh* aMesh, bool aFreeFromMainMemory);

		/** Remove from the video memory. */
		void Remove( Mesh* aMesh );

		/** Send the geometry drawing commands. */
		void Batch( Mesh* aMesh );
		
		/** Create a new texture render target. */
		ScreenRenderTarget* NewScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine );
		
		/** Create a new texture render target. */
		OffscreenRenderTarget* NewOffscreenRenderTarget( int aWidth, int aHeight, bool aColor, bool aDepth );
		
		void RenderScreenQuad( float x0, float y0, float x1, float y1, Texture* aT );
		void RenderScreenQuad( float x0, float y0, float x1, float y1, SimdVector4 aC );
		
	protected:		

		/** Upload a vertex buffer to the video memory. */
		void UploadBuffer( Mesh::VertexBuffer* aBuffer, int aNumVertices, bool aFreeFromMainMemory );
		
		/** Remove a vertex buffer from the video memory. */
		void RemoveBuffer( Mesh::VertexBuffer* aBuffer );
		
	protected:
	
		Program iPhongProgram;
		Program::Parameter iPhongAmbientParameter;
		Program::Parameter iPhongDiffuseParameter;
		Program::Parameter iPhongSpecularParameter;
		Program::Parameter iPhongExponentParameter;
		Program::Parameter iPhongEyePosParameter;
		Program::Parameter iPhongLightPosParameter;
		Program::Parameter iPhongModelTransformParameter;
		Program::Parameter iPhongModelNormalTransformParameter;

		Program iShadowPhongProgram;
		Program::Parameter iShadowPhongAmbientParameter;
		Program::Parameter iShadowPhongDiffuseParameter;
		Program::Parameter iShadowPhongSpecularParameter;
		Program::Parameter iShadowPhongExponentParameter;
		Program::Parameter iShadowPhongEyePosParameter;
		Program::Parameter iShadowPhongLightPosParameter;
		Program::Parameter iShadowPhongDepthBufferParameter;
		Program::Parameter iShadowPhongDepthTransformParameter;
		Program::Parameter iShadowPhongModelTransformParameter;
		Program::Parameter iShadowPhongModelNormalTransformParameter;

		Program iDepthCodeProgram;
		Program::Parameter iDepthCodeModelTransformParameter;		
		
		Program iUVCodeProgram;
		Program::Parameter iUVCodeModelTransformParameter;		
		
		// State
		bool iNullOutput;
		Program* rCurrentProgram;
		Program::Parameter iCurrentModelTransformParameter;
		Program::Parameter iCurrentNormalTransformParameter;
	};
	
}

#endif 
