

#include "RendererGU.h"

#include "GironaLog.h"

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspgu.h>
#include <pspgum.h>

using namespace grn;


static unsigned int __attribute__((aligned(16))) list[262144];

#define BUF_WIDTH (512)
#define SCR_WIDTH (480)
#define SCR_HEIGHT (272)
#define PIXEL_SIZE (4) /* change this if you change to another screenmode */
#define FRAME_SIZE (BUF_WIDTH * SCR_HEIGHT * PIXEL_SIZE)
#define ZBUF_SIZE (BUF_WIDTH SCR_HEIGHT * 2) /* zbuffer seems to be 16-bit? */



RendererGU::RendererGU()
{
	
	// Init state
	iNullOutput = false;
}


RendererGU::~RendererGU()
{
	for ( unsigned int i=0; i<iVideoBuffers.size(); i++ )
		delete iVideoBuffers[i];
	iVideoBuffers.clear();
}

void RendererGU::Activate( Material* aMat )
{

	// Phong material
	if ( PhongMaterial* lPhong = dynamic_cast<PhongMaterial*>(aMat) )
	{
		unsigned int lCol = (0xFF<<24) 
							| ((unsigned int)(lPhong->iDiffuse.iColor[2]*255)<<16)
							| ((unsigned int)(lPhong->iDiffuse.iColor[1]*255)<<8)
							| ((unsigned int)(lPhong->iDiffuse.iColor[0]*255)<<0) ;
							
		sceGuColorMaterial(GU_DIFFUSE);
		sceGuColor( lCol );

		sceGuSpecular(lPhong->iExponent);

		lCol = (0xFF<<24) 
			| ((unsigned int)(lPhong->iAmbient.iColor[2]*255)<<16)
			| ((unsigned int)(lPhong->iAmbient.iColor[1]*255)<<8)
			| ((unsigned int)(lPhong->iAmbient.iColor[0]*255)<<0) ;
			
		sceGuAmbient(lCol);

		iNullOutput = false;

		sceGuLight(0,GU_POINTLIGHT,GU_DIFFUSE_AND_SPECULAR,(ScePspFVector3*)&lPhong->iLightPosition);
		sceGuLightColor(0,GU_AMBIENT,0xffffffff);
		sceGuLightColor(0,GU_DIFFUSE,0xffffffff);
		sceGuLightColor(0,GU_SPECULAR,0xffffffff);
		//sceGuLightAtt(i,0.0f,0.0f,0.0f);

	}
	
	// Depth code material
//	else if ( dynamic_cast<DepthCodeMaterial*>(aMat) )
//	{
//	}
	
	else
	{
		// unsupported material
		iNullOutput = true;
	}

}


void RendererGU::SetModelTransform( const SimdTransform& aT, const SimdTransform& )
{
	ScePspFMatrix4 t;
	aT.getOpenGLMatrix((float*)&t);
	sceGuSetMatrix(GU_MODEL,&t);
}


void RendererGU::Deactivate( Material* )
{
//	if (rCurrentProgram) rCurrentProgram->Deactivate();
}


void RendererGU::Activate( const Camera& aCam ) const
{
	ScePspFMatrix4 t;
	aCam.GetViewMatrix().getValue((float*)&t);
	sceGuSetMatrix(GU_VIEW,&t);
	
	aCam.GetProjectionMatrix().getValue((float*)&t);
	sceGuSetMatrix(GU_PROJECTION,&t);
}


void RendererGU::Batch( Mesh* aMesh )
{
	// Null output canbe caused by an unsupported material
	if (iNullOutput) return;
	
	if ( aMesh->iVertexBuffers[0]->iHandle < iVideoBuffers.size() )
	{
		VideoBuffer *lB = iVideoBuffers[ aMesh->iVertexBuffers[0]->iHandle ];
		sceGumDrawArray(GU_TRIANGLES, lB->iType|GU_TRANSFORM_3D|GU_INDEX_16BIT, aMesh->iNumFaces*3,aMesh->iIndexBuffer->iData,lB->iData);
	}
}


void RendererGU::RenderScreenQuad( float x0, float y0, float x1, float y1, TextureRenderTarget* aT )
{
}


struct QuadVertex
{
	float x,y,z;
};

void RendererGU::RenderScreenQuad( float x0, float y0, float x1, float y1, SimdVector4 aC )
{
	//
	// Doesn work!
	//
	
	sceGumMatrixMode( GU_PROJECTION );
	sceGumLoadIdentity();
	sceGumOrtho(-1.0,1.0,-1.0,1.0,-0.5,0.5);
	sceGumMatrixMode( GU_MODEL);
	sceGumLoadIdentity();
	sceGumMatrixMode( GU_VIEW);
	sceGumLoadIdentity();

	sceGuDisable( GU_DEPTH_TEST );
	sceGuDisable( GU_CULL_FACE );
	sceGuDisable( GU_LIGHTING );
	sceGuEnable( GU_BLEND );												// Turn Blending On
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA,GU_ONE_MINUS_SRC_ALPHA,0,0);		// Blending Function For Translucency Based On Source Alpha Value

	unsigned int lCol =   ((unsigned int)(aC[3]*255)<<24) 
						| ((unsigned int)(aC[2]*255)<<16)
						| ((unsigned int)(aC[1]*255)<<8)
						| ((unsigned int)(aC[0]*255)<<0) ;

	sceGuColor( lCol );

	struct QuadVertex __attribute__((aligned(16))) vertices[6];	
	vertices[0].x = x0;
	vertices[0].y = y0;
	vertices[0].z = 0;
	
	vertices[1].x = x1;
	vertices[1].y = y0;
	vertices[1].z = 0;

	vertices[2].x = x1;
	vertices[2].y = y1;
	vertices[2].z = 0;

	vertices[3].x = x0;
	vertices[3].y = y0;
	vertices[3].z = 0;
	
	vertices[4].x = x1;
	vertices[4].y = y1;
	vertices[4].z = 0;

	vertices[5].x = x0;
	vertices[5].y = y1;
	vertices[5].z = 0;
	
	sceGumDrawArray(GU_TRIANGLES,GU_VERTEX_32BITF|GU_TRANSFORM_3D,6,0,vertices);

	sceGuDisable( GU_BLEND );		// Turn Blending Off
	sceGuEnable( GU_LIGHTING );
	sceGuEnable( GU_CULL_FACE );	
	sceGuEnable( GU_DEPTH_TEST );
	
}



typedef struct Vertex
{
 	float x,y,z;
} Vertex;

typedef struct NormalVertex
{
	float nx,ny,nz;
 	float x,y,z;
} NormalVertex;

typedef struct TexVertex
{
	float u,v;
 	float x,y,z;
} TexVertex;

typedef struct TexNormalVertex
{
	float u,v;
	float nx,ny,nz;
 	float x,y,z;
} TexNormalVertex;


void RendererGU::Upload( Mesh* aMesh, bool aFreeFromMainMemory)
{
	float* lPos = 0;
	float* lNorm = 0;
	float* lTex = 0;
	
	for ( unsigned int i=0; i<aMesh->iVertexBuffers.size(); i++ )
		if ( aMesh->iVertexBuffers[i]->iSemantic == Mesh::VertexBuffer::ePosition ) lPos = (float*)aMesh->iVertexBuffers[i]->iData;
		else if ( aMesh->iVertexBuffers[i]->iSemantic == Mesh::VertexBuffer::eNormal ) lNorm = (float*)aMesh->iVertexBuffers[i]->iData;
		else if ( aMesh->iVertexBuffers[i]->iSemantic == Mesh::VertexBuffer::eTexCoord0 ) lTex = (float*)aMesh->iVertexBuffers[i]->iData;
		
	if ( lPos && !lNorm && !lTex )
	{
		int lH = iVideoBuffers.size();
		aMesh->iVertexBuffers[0]->iHandle = lH;

		iVideoBuffers.push_back( new VideoBuffer());
		iVideoBuffers[lH]->iType = GU_VERTEX_32BITF;
		iVideoBuffers[lH]->iData = malloc(sizeof(Vertex)*aMesh->iNumVertices);
		Vertex* lB = (Vertex*)iVideoBuffers[lH]->iData;
		for (int i=0;i<aMesh->iNumVertices; i++)
		{
			lB->x = lPos[0];
			lB->y = lPos[1];
			lB->z = lPos[2];
			lPos += 3;
			lB++;
		}		
	}
	
	else if ( lPos && lNorm && !lTex )
	{
		int lH = iVideoBuffers.size();
		aMesh->iVertexBuffers[0]->iHandle = lH;

		iVideoBuffers.push_back( new VideoBuffer() );
		iVideoBuffers[lH]->iType = GU_VERTEX_32BITF | GU_NORMAL_32BITF;
		iVideoBuffers[lH]->iData = malloc(sizeof(NormalVertex)*aMesh->iNumVertices);
		NormalVertex* lB = (NormalVertex*)iVideoBuffers[lH]->iData;

		for (int i=0;i<aMesh->iNumVertices; i++)
		{
			lB->x = lPos[0];
			lB->y = lPos[1];
			lB->z = lPos[2];
			lB->nx = lNorm[0];
			lB->ny = lNorm[1];
			lB->nz = lNorm[2];
			lPos += 3;
			lNorm += 3;
			lB++;
		}
	}

	else grn::force::log("Unsupported vertex format.");
	
}


void RendererGU::Remove( Mesh* aMesh )
{
	// replacing by empty will call the destructor in previous and not change any position for other buffers.
	delete iVideoBuffers[aMesh->iVertexBuffers[0]->iHandle];
	iVideoBuffers[aMesh->iVertexBuffers[0]->iHandle] = 0;
}


//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////



ScreenRenderTarget* RendererGU::NewScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine )
{
	return new GUScreenRenderTarget(aX,aY,aWidth,aHeight,aEngine);
}
		

OffscreenRenderTarget* RendererGU::NewOffscreenRenderTarget( int aWidth, int aHeight, bool aColor=true, bool aDepth=true )
{
	return new GUOffscreenRenderTarget(aWidth,aHeight,aColor,aDepth);
}


GUTextureRenderTarget::GUTextureRenderTarget( int aWidth, int aHeight, bool aColor, bool aDepth )
	: TextureRenderTarget(aWidth,aHeight)
	, iTexture(0)
{
}


void GUTextureRenderTarget::Activate()
{
	/*
	iTexture->BeginCapture();
	glViewport( 0, 0, iWidth, iHeight );
*/
}


void GUTextureRenderTarget::Deactivate()
{
	//iTexture->EndCapture();
}


void GUTextureRenderTarget::Present()
{
}


void GUTextureRenderTarget::Bind()
{ 
//	iTexture->Bind(); 
}


void GUTextureRenderTarget::Clear()
{
	/*
	Activate();
	glClearColor(1.0,1.0,1.0,1.0f);
	glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
	Deactivate();
	*/
}


//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////


GUScreenRenderTarget::GUScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine )
	: ScreenRenderTarget(aX,aY,aWidth,aHeight,aEngine)
{
	sceGuStart(GU_DIRECT,list);
	sceGuDrawBuffer(GU_PSM_8888,(void*)0,BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,(void*)0x88000,BUF_WIDTH);
	sceGuDepthBuffer((void*)0x110000,BUF_WIDTH);
	sceGuOffset(2048 - (SCR_WIDTH/2),2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048,2048,SCR_WIDTH,SCR_HEIGHT);
	sceGuDepthRange(0xc350,0x2710);
	sceGuScissor(0,0,SCR_WIDTH,SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuDepthFunc(GU_GEQUAL);
	sceGuEnable(GU_DEPTH_TEST);
	sceGuFrontFace(GU_CCW);
	sceGuShadeModel(GU_SMOOTH);
	sceGuDisable(GU_CULL_FACE);
	sceGuEnable(GU_LIGHTING);
	sceGuEnable(GU_LIGHT0);
	sceGuFinish();
	sceGuSync(0,0);

	sceDisplayWaitVblankStart();
	sceGuDisplay(GU_TRUE);
}


void GUScreenRenderTarget::Activate()
{
	//__psp_log("------------------- start2 \n");
	//sceGuStart(GU_DIRECT,list);
	//log("[engine] Activate screen: start.");
}


void GUScreenRenderTarget::Deactivate()
{
	//__psp_log("------------------- finish2 \n");
	//sceGuFinish();
	//log("[engine] deactivate screen: finish.");
	
}

void GUScreenRenderTarget::Clear()
{
	// clear screen
	//__psp_log("------------------- start \n");
	sceGuStart(GU_DIRECT,list);
	sceGuClearColor(0xff000000);
	sceGuClearDepth(0);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
	//sceGuFinish();

}

void GUScreenRenderTarget::Present()
{
	
	ScreenRenderTarget::Present();
}



