#ifndef GRN_RENDERER_GU_H_
#define GRN_RENDERER_GU_H_


#include "Renderer.h"

#include <vector>

namespace grn
{
	
	class GUOffscreenRenderTarget : public OffscreenRenderTarget
	{
	public:
		GUOffscreenRenderTarget( int aWidth, int aHeight, bool aColor=true, bool aDepth=true );
			
		virtual void Activate();
		virtual void Deactivate();
		virtual void Present();

		virtual void Bind();
		virtual void Clear();
		
		RenderTexture* iTexture;
		
	};
	
	
	class GUScreenRenderTarget : public ScreenRenderTarget
	{
	public:
		GUScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine );

		virtual void Activate();
		virtual void Deactivate();
		virtual void Present();
		
		virtual void Clear();
	};
	

	class RendererGU
		: public Renderer
	{
	public:
		RendererGU();
		~RendererGU();
		
		void Activate( const Camera& aCam ) const;
		void Activate( Material* aMat );
		void Deactivate( Material* aMat );
		void SetModelTransform( const SimdTransform&, const SimdTransform& );

		void Render( const AABB& aB ) const {};
		void RenderInverted( const AABB& aB ) const {};
		
		/** Upload all buffers to the video memory. */
		void Upload( Mesh* aMesh, bool aFreeFromMainMemory);

		/** Remove from the video memory. */
		void Remove( Mesh* aMesh );

		/** Send the geometry drawing commands. */
		void Batch( Mesh* aMesh );
		
		/** Create a new texture render target. */
		ScreenRenderTarget* NewScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine );
		
		/** Create a new texture render target. */
		OffscreenRenderTarget* NewOffscreenRenderTarget( int aWidth, int aHeight, bool aColor, bool aDepth );
		
		void RenderScreenQuad( float x0, float y0, float x1, float y1, TextureRenderTarget* aT );
		void RenderScreenQuad( float x0, float y0, float x1, float y1, SimdVector4 aC );
		
	protected:
	
		// State
		bool iNullOutput;
		
		/** "Video" buffer definition. */
		class VideoBuffer
		{
		public:
			VideoBuffer() : iData(0) {};
			~VideoBuffer() { if (iData) free(iData); iData = 0; }
			void* iData;
			unsigned int iType;
		};
		
		/** Buffers prepared to be sent directly to the PSP. */
		std::vector<VideoBuffer*> iVideoBuffers;
	};
	
}

#endif 
