#ifndef GRN_RENDERER_H_
#define GRN_RENDERER_H_


#include "datatypes/AABB.h"
#include "datatypes/Camera.h"
#include "datatypes/Material.h"
#include "datatypes/Mesh.h"
#include "datatypes/RenderTarget.h"

namespace grn
{

	class Engine;

	class Renderer
	{
	public:

		virtual ~Renderer() {}

		virtual void Activate( const Camera& aCam ) const = 0;
		virtual void Activate( Material* aMat ) = 0;
		virtual void Deactivate( Material* aMat ) = 0;
		virtual void SetModelTransform( const SimdTransform&, const SimdTransform& ) = 0;

		virtual void Render( const AABB& aB ) const = 0;
		virtual void RenderInverted( const AABB& aB ) const = 0;

		/** Upload all buffers to the video memory. */
		virtual void Upload( Mesh* aMesh, bool aFreeFromMainMemory ) = 0;

		/** Remove from the video memory. */
		virtual void Remove( Mesh* aMesh ) = 0;

		/** Send the geometry drawing commands. */
		virtual void Batch( Mesh* aMesh ) = 0;
		
		/** Create a new screen render target. */
		virtual ScreenRenderTarget* NewScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine ) = 0;
		
		/** Create a new texture render target. */
		virtual OffscreenRenderTarget* NewOffscreenRenderTarget( int aWidth, int aHeight, bool aColor=true, bool aDepth=true ) = 0;

		virtual void RenderScreenQuad( float x0, float y0, float x1, float y1, Texture* aT ) = 0;
		virtual void RenderScreenQuad( float x0, float y0, float x1, float y1, SimdVector4 aC ) = 0;
		
	};
	
}

#endif 
