
#include "Export.h"

#include "Pipeline.h"

#include <sstream>


using namespace grn;

IMPLEMENT_MODULE(Export);


Export::Export( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	rInput = 0;
	rPipelinePad = 0;
	iExportedType = 0;
	bzero(iName,sizeof(char)*GRN_EXPORT_MAX_NAME_LENGTH);
	UpdateSelectedType();	
}


void Export::Run()
{
	// Emit outside this pipeline if pipeline has a parent
	if (rPipeline->GetParent())
		while (rInput->hasData())
			rPipeline->GetParent()->Emit( rPipelinePad, rInput->getGenericData( true ) );
}


void Export::UpdateSelectedType()
{
	if (rInput)
	{
		// Removes, unlinks and deletes
		RemovePad( rInput );
		rPipeline->RemovePad( rPipelinePad );
	}
	
	DataType lType = rPipeline->FindDataType(iExportedType);
	rInput = new Module::InputPad( lType, "ExportedData", this );
	addPad(rInput);

	rPipelinePad = new Module::OutputPad( lType, std::string(iName)+"::"+lType );
	rPipeline->addPad(rPipelinePad);
}


void Export::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);

	s<< std::string(iName) <<std::endl;
	s<< rPipeline->FindDataType(iExportedType) <<std::endl;
}


void Export::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	s>> iName;
	
	DataType l;
	s>> l;

	iExportedType = rPipeline->GetDataTypeID(l);

	UpdateSelectedType();
}

