
#include "If.h"

#include "Pipeline.h"

#include <sstream>


using namespace grn;


IMPLEMENT_MODULE(If);

If::If( Pipeline* aParent )
: Module(aParent)
{
	setSize(80,40);
	iX = 300;
	iY = 150;
	iImportedType = 0;
	rOutput = 0;
	rInputTrue = 0;
	rInputFalse = 0;
	riIndex = new Module::tInputPad<grn::Boolean>( "Condition", this );
	UpdateSelectedType();
}


void If::Run()
{
	while (AllInputsHaveData())
	{
		bool Selected = riIndex->getData(true)->iValue;
		Pointer<DataObject> lTrue = rInputTrue->getGenericData(true);
		Pointer<DataObject> lFalse = rInputFalse->getGenericData(true);
		if (Selected) rPipeline->Emit( rOutput, lTrue );
		else rPipeline->Emit( rOutput, lFalse );
	}
}


void If::UpdateSelectedType()
{
	if (rOutput)
	{
		// Removes, unlinks and deletes
		RemovePad( rOutput );
		RemovePad( rInputTrue );
		RemovePad( rInputFalse );
	}
	
	DataType lType = rPipeline->FindDataType(iImportedType);
	rOutput = new Module::OutputPad( lType, "Selected" );
	addPad(rOutput);
	rInputTrue = new Module::InputPad( lType, "True", this );
	addPad(rInputTrue);
	rInputFalse = new Module::InputPad( lType, "False", this );
	addPad(rInputFalse);
}


void If::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s<< rPipeline->FindDataType(iImportedType) <<std::endl;
	
}


void If::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	DataType l;
	s>> l;
	iImportedType = rPipeline->GetDataTypeID(l);
	
	UpdateSelectedType();
}

