
#include "Import.h"

#include "Pipeline.h"


#include <sstream>


using namespace grn;


IMPLEMENT_MODULE(Import);

Import::Import( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	iImportedType = 0;
	iAccumulate = 0;
	rOutput = 0;
	rPipelinePad = 0;
	bzero(iName,sizeof(char)*GRN_MAX_NAME_LENGTH);
	UpdateSelectedType();
}


void Import::Setup()
{
	iAccumulated.clear();
}


void Import::Init()
{
	// Emit data coming from outside this pipeline
	
	if (!iAccumulate)
		while (rPipelinePad->hasData())
			rPipeline->Emit( rOutput, rPipelinePad->getGenericData( true ) );
	else
	{
		while (rPipelinePad->hasData())
			iAccumulated.push_back(rPipelinePad->getGenericData(true));

		for ( std::vector< Pointer<DataObject> >::iterator it = iAccumulated.begin()
			; it !=  iAccumulated.end()
			; it++ )
			rPipeline->Emit( rOutput, *it );
	}
		
}


void Import::UpdateSelectedType()
{
	if (rOutput)
	{
		// Removes, unlinks and deletes
		RemovePad( rOutput );
		rPipeline->RemovePad( rPipelinePad );
	}
	
	DataType lType = rPipeline->FindDataType(iImportedType);
	rOutput = new Module::OutputPad( lType, "ImportedData" );
	addPad(rOutput);

	rPipelinePad = new Module::InputPad( lType, std::string(iName)+"::"+lType, rPipeline );
	rPipeline->addPad(rPipelinePad);
}


void Import::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s<< std::string(iName) <<std::endl;
	s<< rPipeline->FindDataType(iImportedType) <<std::endl;
	s<< iAccumulate <<std::endl;
	
}


void Import::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	s>> iName;

	DataType l;
	s>> l;
	iImportedType = rPipeline->GetDataTypeID(l);
	
	s>> iAccumulate;

	UpdateSelectedType();
}

