
#include "MakeGroup.h"

#include "Pipeline.h"
#include "GironaLog.h"


using namespace grn;


IMPLEMENT_MODULE(MakeGroup);


MakeGroup::MakeGroup( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,80);
	iX = 300;
	riObjects = new Module::tInputPad<grn::Object>( "Objects", this );
	riCameras = new Module::tInputPad<grn::Camera>( "Cameras", this );
	riSpotLights = new Module::tInputPad<grn::SpotLight>( "SpotLights", this );
	riEnvs = new Module::tInputPad<grn::PhysicsEnvironment>( "PhysicsEnvironments", this );
	roGroups = new Module::tOutputPad<grn::Group>( "Groups", this );
}


void MakeGroup::Init()
{
	iGroup = new Group();
	iGroup->iCollections.push_back( Group::Collection("Objects") );
	iGroup->iCollections.push_back( Group::Collection("Cameras") );
	iGroup->iCollections.push_back( Group::Collection("Spotlights") );
	iGroup->iCollections.push_back( Group::Collection("Envs") );
}


void MakeGroup::Run()
{
	// Collect all data	
	while ( riObjects->hasData() ) iGroup->iCollections[0].iObjects.push_back( riObjects->getGenericData(true) );
	while ( riCameras->hasData() ) iGroup->iCollections[1].iObjects.push_back( riCameras->getGenericData(true) );
	while ( riSpotLights->hasData() ) iGroup->iCollections[2].iObjects.push_back( riSpotLights->getGenericData(true) );
	while ( riEnvs->hasData() ) iGroup->iCollections[3].iObjects.push_back( riEnvs->getGenericData(true) );
}


bool MakeGroup::Eod()
{
	bool lConsumed = false;
	
	while ( AllInputsHaveEOD() )
	{
		rPipeline->Emit(roGroups, iGroup);

		log("[MakeGroup::Eod] Group emitted : ");
		log("[MakeGroup::Eod]  collection 1 : ",(int)iGroup->iCollections[0].iObjects.size());
		log("[MakeGroup::Eod]  collection 2 : ",(int)iGroup->iCollections[1].iObjects.size());
		log("[MakeGroup::Eod]  collection 3 : ",(int)iGroup->iCollections[2].iObjects.size());
		log("[MakeGroup::Eod]  collection 4 : ",(int)iGroup->iCollections[3].iObjects.size());

		// Reinitialize the group in progress
		Init();
		
		// Clear EODs
		for ( std::vector<InputPad*>::iterator it = iInputs.begin()
			; it != iInputs.end()
			; it++ )
			(*it)->getGenericData(true);
		
		lConsumed = true;
	}
	
	return lConsumed;
}

