
#include "MarkGroup.h"

#include "Pipeline.h"

#include <sstream>


using namespace grn;


IMPLEMENT_MODULE(MarkGroup);

MarkGroup::MarkGroup( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	iImportedType = 0;
	rOutput = 0;
	rInput = 0;
	UpdateSelectedType();
}


void MarkGroup::Run()
{
	// Emit data coming from input interleaving the EOD signal
	
	while (rInput->hasData())
	{
		rPipeline->Emit( rOutput, rInput->getGenericData( true ) );
		rPipeline->EmitEOD( rOutput );
	}
		
}


void MarkGroup::UpdateSelectedType()
{
	if (rOutput)
	{
		// Removes, unlinks and deletes
		RemovePad( rOutput );
		RemovePad( rInput );
	}
	
	DataType lType = rPipeline->FindDataType(iImportedType);
	rOutput = new Module::OutputPad( lType, "MarkedData" );
	addPad(rOutput);
	rInput = new Module::InputPad( lType, "RawData", this );
	addPad(rInput);
}


void MarkGroup::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s<< rPipeline->FindDataType(iImportedType) <<std::endl;
	
}


void MarkGroup::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	DataType l;
	s>> l;
	iImportedType = rPipeline->GetDataTypeID(l);
	
	UpdateSelectedType();
}

