
#include "Replicate.h"

#include "Pipeline.h"

#include <sstream>


using namespace grn;


IMPLEMENT_MODULE(Replicate);

Replicate::Replicate( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,40);
	iX = 300;
	iY = 150;
	iSourceType = 0;
	iTimesType = 0;
	rOutput = 0;
	riSource = 0;
	riTimes = 0;
	UpdateSelectedTypes();
}


void Replicate::Run()
{
	while (AllInputsHaveData())
	{
		// Get source but don't remove
		Pointer<DataObject> lSource = riSource->getGenericData(false);
		// Consume one from "times"
		riTimes->getGenericData(true);
		rPipeline->Emit( rOutput, lSource );
	}
}


void Replicate::UpdateSelectedTypes()
{
	if (rOutput)
	{
		// Removes, unlinks and deletes
		RemovePad( rOutput );
		RemovePad( riSource );
		RemovePad( riTimes );
	}
	
	DataType lType = rPipeline->FindDataType(iSourceType);
	rOutput = new Module::OutputPad( lType, "Replicated" );
	addPad(rOutput);
	riSource = new Module::InputPad( lType, "Source", this );
	addPad(riSource);

	lType = rPipeline->FindDataType(iTimesType);
	riTimes = new Module::InputPad( lType, "Times", this );
	addPad(riTimes);
}


void Replicate::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s<< rPipeline->FindDataType(iSourceType) <<std::endl;
	s<< rPipeline->FindDataType(iTimesType) <<std::endl;
	
}


void Replicate::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	DataType l;
	s>> l;
	iSourceType = rPipeline->GetDataTypeID(l);
	s>> l;
	iTimesType = rPipeline->GetDataTypeID(l);
	
	UpdateSelectedTypes();
}

