
#include "SequenceSelect.h"

#include "Pipeline.h"

#include <sstream>


using namespace grn;


IMPLEMENT_MODULE(SequenceSelect);

SequenceSelect::SequenceSelect( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	iImportedType = 0;
	rOutput = 0;
	rInput = 0;
	riIndex = new Module::tInputPad<grn::Integer>( "Index", this );
	UpdateSelectedType();
}


void SequenceSelect::Run()
{
	
	while (riIndex->hasData())
	{
		int Selected = riIndex->getData(true)->iValue-1;
		if ( (Selected>=0) && ((int)rInput->iDataQueue.size()>Selected) )
		{
			InputPad::DataQueue::iterator it = rInput->iDataQueue.begin();
			for (int i=0;i<Selected;i++) it++;
			rPipeline->Emit( rOutput, *it );
		}
	}
		
}


void SequenceSelect::UpdateSelectedType()
{
	if (rOutput)
	{
		// Removes, unlinks and deletes
		RemovePad( rOutput );
		RemovePad( rInput );
	}
	
	DataType lType = rPipeline->FindDataType(iImportedType);
	rOutput = new Module::OutputPad( lType, "Selected" );
	addPad(rOutput);
	rInput = new Module::InputPad( lType, "Sequence", this );
	addPad(rInput);
}


void SequenceSelect::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s<< rPipeline->FindDataType(iImportedType) <<std::endl;
	
}


void SequenceSelect::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	DataType l;
	s>> l;
	iImportedType = rPipeline->GetDataTypeID(l);
	
	UpdateSelectedType();
}

