
#include "Fade.h"

#include "Pipeline.h"
#include "engine/Engine.h"
#include "engine/Renderer.h"


IMPLEMENT_MODULE(Fade); 

using namespace grn;


Fade::Fade( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,80);
	iX = 300;
	riSource = new Module::tInputPad<grn::RenderTarget>( "SourceRenderTarget", this );
	riColor = new Module::tInputPad<grn::Vector>( "Color", this );
	riIntensity = new Module::tInputPad<grn::Real>( "Intensity", this );
	roTarget = new Module::tOutputPad<grn::RenderTarget>( "RenderTarget", this );
}


void Fade::Link(Engine* aEngine)
{
	rEngine = aEngine;
}


void Fade::Run()
{
	while ( AllInputsHaveData() )
	{
		// Fetch render target, and remove
		Pointer<RenderTarget> lTarget = riSource->getData( true );

		lTarget->Activate();

		// Fetch color and intensity
		Pointer<Vector> lColor = riColor->getData( true );
		Pointer<Real> lIntensity = riIntensity->getData( true );

		rEngine->GetRenderer()->RenderScreenQuad(-1.0,-1.0,1.0,1.0, SimdVector4(lColor->iValue[0],lColor->iValue[1],lColor->iValue[2],lIntensity->iValue) );
		
		lTarget->Deactivate();

		// emit the render target
		rPipeline->Emit( roTarget, lTarget );
	}
}
