
#include "GetRenderTexture.h"

#include "Pipeline.h"
#include "GironaLog.h"
#include "engine/Engine.h"
#include "engine/Renderer.h"
#include "datatypes/Camera.h"


IMPLEMENT_MODULE(GetRenderTexture); 

using namespace grn;


GetRenderTexture::GetRenderTexture( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,80);
	iX = 300;
	riTarget = new Module::tInputPad<grn::RenderTarget>( "RenderTarget", this );
	roTarget = new Module::tOutputPad<grn::RenderTarget>( "RenderTarget", this );
	roTexture = new Module::tOutputPad<grn::Texture>( "Texture", this );
}


void GetRenderTexture::Link( Engine* aEngine )
{
	rEngine = aEngine;
}


void GetRenderTexture::Run()
{
	
	while ( AllInputsHaveData() )
	{
		// Fetch render target, and remove
		Pointer<OffscreenRenderTarget> lTarget = dynamic_cast<OffscreenRenderTarget*>((RenderTarget*)riTarget->getData( true ));

		// Fetch object and remove
		Pointer<Texture> lIn = lTarget->GetColorTexture();

		if (lIn)
			// Emit the texture
			rPipeline->Emit( roTexture, lIn );
		else
			grn::force::log("[error] RenderTarget is not offscreen in GetRenderTexture.");
		
		// emit the render target
		rPipeline->Emit( roTarget, lTarget );
	}
}
