
#include "MeshRender.h"

#include "GironaLog.h"
#include "Pipeline.h"
#include "datatypes/Camera.h"
#include "engine/Renderer.h"

using namespace grn;

IMPLEMENT_MODULE(MeshRender);


MeshRender::MeshRender( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,110);
	iX = 300;
	riMesh = new Module::tInputPad<grn::Mesh>( "Mesh", this );
	riMaterial = new Module::tInputPad<grn::Material>( "Material", this );
	riTransform = new Module::tInputPad<grn::Transform>( "Transform", this );
	riCamera = new Module::tInputPad<grn::Camera>( "Camera", this );
	riTarget = new Module::tInputPad<grn::RenderTarget>( "RenderTarget", this );
	roTarget = new Module::tOutputPad<grn::RenderTarget>( "RenderTarget", this );
	
	rRenderer = 0;
}


void MeshRender::Link( Engine* aEngine )
{
	rRenderer = aEngine->GetRenderer();
}


void MeshRender::Run()
{

	if ( AllInputsHaveData() )
	{
		// Fetch render target, don't remove
		Pointer<RenderTarget> lTarget = riTarget->getData(false);
		lTarget->Activate();
		
		while ( riMesh->hasData() && riMaterial->hasData() && riTransform->hasData() )
		{			
			// Fetch camera
			Pointer<Camera> lCam = riCamera->getData(false);
			rRenderer->Activate(*lCam);
	
			// Fetch material and remove
			Pointer<Material> lMaterial = riMaterial->getData(true);
			lMaterial->SetCameraParameter("View",lCam);
			rRenderer->Activate( lMaterial );
	
			// Fetch transform and remove
			Pointer<Transform> lTransform = riTransform->getData(true);
			SimdTransform lT(lTransform->GetTransform());
			SimdTransform lNormals(lT.getRotation());
			rRenderer->SetModelTransform( lT, lNormals );
			
			// Fetch mesh and remove
			Pointer<Mesh> lInMesh = riMesh->getData(true);
			
			rRenderer->Batch(lInMesh);
			
			rRenderer->Deactivate( lMaterial );
	
		}
	
		lTarget->Deactivate();
	}
	
}


bool MeshRender::Eod()
{	
	bool lConsumed = false;
	
	// Consume any EOD in meaningless inputs
	while (riMaterial->hasEOD()) riMaterial->getData(true);
	while (riTransform->hasEOD()) riTransform->getData(true);

	if (riTarget->hasData())
	{
		if (riMesh->hasEOD()) 
		{
			lConsumed = true;				

			// Consume the mesh EOD
			riMesh->getData(true);
			
			// Consume camera and remove EOD if any (it should be...)
			if (riCamera->hasData()) riCamera->getData(true);
			while (riCamera->hasEOD()) riCamera->getData(true);
			
			// Consume RenderTarget and remove EOD if any
			Pointer<RenderTarget> lRT = riTarget->getData(true);
			while (riTarget->hasEOD()) riTarget->getData(true);
			
			rPipeline->Emit(roTarget,lRT);
		}

	}
	
	return lConsumed;
}


