
#include "PhysicsModifier.h"

#include "Pipeline.h"

#include "GironaLog.h"


using namespace grn;

IMPLEMENT_MODULE(PhysicsModifier);


PhysicsModifier::PhysicsModifier( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,40);
	iX = 300;
	iY = 150;
	riObjects = new Module::tInputPad<grn::Object>( "InputObjects", this );
	riVector = new Module::tInputPad<grn::Vector>( "Vector", this );
	iProperty = 0;
}


void PhysicsModifier::Run()
{
	while ((riVector->hasData())&&(riObjects->hasData()))
	{
		Pointer<Object> lInObj = riObjects->getData(true);
		Pointer<Vector> lV = riVector->getData(true);

		//lInObj->iTransform->iTransform.translate( SimdVector3(0.0,0.0,1.0f) );		
		assert (lInObj->iPhysicsController);
		if (lInObj->iPhysicsController)
			switch (iProperty)
			{
			case 0: lInObj->iPhysicsController->SetLinearVelocity( lV->iValue.x(), lV->iValue.y(), lV->iValue.z(), iLocal ); break;
			case 1: lInObj->iPhysicsController->SetAngularVelocity( lV->iValue.x(), lV->iValue.y(), lV->iValue.z(), iLocal ); break;
			case 2: lInObj->iPhysicsController->ApplyForce( lV->iValue.x(), lV->iValue.y(), lV->iValue.z(), iLocal ); break;
			case 3: lInObj->iPhysicsController->ApplyTorque( lV->iValue.x(), lV->iValue.y(), lV->iValue.z(), iLocal ); break;
			case 4: 
			{
				//log4("   setting orientation : ",lV->iValue.x(), lV->iValue.y(), lV->iValue.z());
				// Set orientation (always world orientation) for both the physics object and the geometric one
				lInObj->SetWorldOrientation( lV->iValue.x(), lV->iValue.y(), lV->iValue.z() );
				SimdQuaternion aQ( lV->iValue.x(), lV->iValue.y(), lV->iValue.z() );
				lInObj->iPhysicsController->setOrientation( aQ.x(), aQ.y(), aQ.z(), aQ[3] ); break;
				break;
			}
			default: break;
			}
	}
}


void PhysicsModifier::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s << iProperty << std::endl;
	s << iLocal << std::endl;
}


void PhysicsModifier::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	s>>iProperty;
	s>>iLocal;
}
