
#include "VariableVector.h"

#include "Pipeline.h"



using namespace grn;

IMPLEMENT_MODULE(VariableVector);


VariableVector::VariableVector( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	roVector = new Module::tOutputPad<grn::Vector>( "Vector", this );
	riVector = new Module::tInputPad<grn::Vector>( "NewValue", this );
	ivX = ivXinitial = 0.0f;
	ivY = ivYinitial = 0.0f;
	ivZ = ivZinitial = 0.0f;
}


void VariableVector::Setup()
{
	ivX = ivXinitial;
	ivY = ivYinitial;
	ivZ = ivZinitial;
}


void VariableVector::Init()
{
	Pointer<Vector> lV = new Vector();
	lV->iValue.setX( ivX );
	lV->iValue.setY( ivY );
	lV->iValue.setZ( ivZ );
	rPipeline->Emit( roVector, lV );
}


void VariableVector::Run()
{
	// Process all input vectors. It only stores the last one received.
	// Don't emit. Keep the value for the next iteration.
	while (AllInputsHaveData())
	{
		Pointer<Vector> lNewValue = riVector->getData(true);
		ivX = lNewValue->iValue.x();
		ivY = lNewValue->iValue.y();
		ivZ = lNewValue->iValue.z();
	}
}


void VariableVector::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s << ivXinitial << " " << ivYinitial << " " << ivZinitial << std::endl;
}


void VariableVector::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	s>>ivXinitial;
	s>>ivYinitial;
	s>>ivZinitial;
}
