
#include "AddToPhysicsEnv.h"

#include "Pipeline.h"

//#include "blender/blenkernel/BKE_material.h"

#include "PHY_IMotionState.h"
#include "SimdMinMax.h"
//#include "MEM_guardedalloc.h"
//#include "kernel/gen_system/GEN_HashedPtr.h"
//#include "kernel/gen_system/GEN_Map.h"

#include "PHY_Pro.h"

#include "CcdPhysicsEnvironment.h"
#include "CcdPhysicsController.h"

#include "CollisionShapes/BoxShape.h"
#include "CollisionShapes/SphereShape.h"
#include "CollisionShapes/ConeShape.h"
#include "CollisionShapes/ConvexShape.h"
#include "CollisionShapes/CylinderShape.h"
#include "CollisionShapes/MultiSphereShape.h"
#include "CollisionShapes/ConvexHullShape.h"
#include "CollisionShapes/TriangleMesh.h"
#include "CollisionShapes/TriangleMeshShape.h"


using namespace grn;

IMPLEMENT_MODULE(AddToPhysicsEnv);


typedef enum {
	KX_BOUNDBOX,
	KX_BOUNDSPHERE,
	KX_BOUNDCYLINDER,
	KX_BOUNDCONE,
	KX_BOUNDMESH,
	KX_BOUNDPOLYTOPE
} KX_BoundBoxClass;


struct KX_BoxBounds
{
	float m_center[3];
	float m_extends[3];
};


/* Cone/Cylinder */
struct KX_CBounds
{
	float m_radius;
	float m_height;
};


struct KX_ObjectProperties
{
	bool	m_dyna;
	double	m_radius;
	bool	m_angular_rigidbody;
	bool	m_in_active_layer;
	bool	m_ghost;
	class KX_GameObject*	m_dynamic_parent;
	bool	m_isactor;
	bool	m_concave;
	bool	m_isdeformable;
	KX_BoundBoxClass	m_boundclass;
	union {
		KX_BoxBounds	box;
		KX_CBounds	c;
	} m_boundobject;
};


static PHY_MaterialProps g_materialProps = {
	1.0,    // restitution
	2.0,    // friction 
	0.0,    // fh spring constant
	0.0,    // fh damping
	0.0,    // fh distance
	false   // sliding material?
};


void ConvertBulletObject( grn::Object* gameobj
						, PHY_IPhysicsEnvironment* aEnvironment
						, PHY_ShapeProps* shapeprops
						, grn::PhysicsProperties* lProperties
						, KX_ObjectProperties*	objprop );
								
																

AddToPhysicsEnv::AddToPhysicsEnv( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,70);
	iX = 500;
	riObject = new Module::tInputPad<Object>( "Object", this );
	riProps = new Module::tInputPad<PhysicsProperties>( "PhysicsProperties", this );
	riEnv = new Module::tInputPad<PhysicsEnvironment>( "PhysicsEnvironment", this );
}


void AddToPhysicsEnv::Run()
{
	if ( riEnv->hasData() )
	{
		Pointer<PhysicsEnvironment> lEnv = riEnv->getData(false);
		
		while ( riObject->hasData() && riProps->hasData() )
		{	
			// Fetch object
			Pointer<PhysicsProperties> lProperties = riProps->getData(true);
			Pointer<Object> lObject = riObject->getData(true);
	
			CreatePhysicsObjectNew( lObject, lProperties, lEnv->iPhysicsEnvironment );
			
			// Add raw data to environment to allow serialisation
			lEnv->iRawData.push_back( std::pair< Pointer<Object>, Pointer<PhysicsProperties> > ( lObject, lProperties ) );
		}
	}
		
}


bool AddToPhysicsEnv::Eod()
{
	if ( riObject->hasEOD() && riProps->hasEOD() )
		while ( riEnv->hasData() ) riEnv->getData( true );
	return Module::Eod();
}

/*
static PHY_MaterialProps *CreateMaterialFromBlenderObject( struct ::Object* blenderobject )
{
	PHY_MaterialProps *materialProps = new PHY_MaterialProps();
	MT_assert(materialProps && "Create physics material properties failed");
		
	::Material* blendermat = give_current_material(blenderobject, 0);
		
	if (blendermat)
	{
		MT_assert(0.0f <= blendermat->reflect && blendermat->reflect <= 1.0f);
	
		materialProps->m_restitution = blendermat->reflect;
		materialProps->m_friction = blendermat->friction;
		materialProps->m_fh_spring = blendermat->fh;
		materialProps->m_fh_damping = blendermat->xyfrict;
		materialProps->m_fh_distance = blendermat->fhdist;
		materialProps->m_fh_normal = (blendermat->dynamode & MA_FH_NOR) != 0;
	}
	else {
		*materialProps = g_materialProps;
	}
	
	return materialProps;
}
*/

static PHY_ShapeProps *CreateShapePropsFromBlenderObject( grn::PhysicsProperties* lProperties )
{
	PHY_ShapeProps *shapeProps = new PHY_ShapeProps();
	assert(shapeProps);
		
	shapeProps->m_mass = lProperties->iMass;
	shapeProps->m_inertia = lProperties->iInertia;
	
	assert(0.0f <= lProperties->iLinearDrag && lProperties->iLinearDrag <= 1.0f);
	assert(0.0f <= lProperties->iAngularDrag && lProperties->iAngularDrag <= 1.0f);
	
	shapeProps->m_lin_drag = 1.0 - lProperties->iLinearDrag;
	shapeProps->m_ang_drag = 1.0 - lProperties->iAngularDrag;
	
	shapeProps->m_friction_scaling[0] = lProperties->iAnisotropicFrictionScaling[0]; 
	shapeProps->m_friction_scaling[1] = lProperties->iAnisotropicFrictionScaling[1];
	shapeProps->m_friction_scaling[2] = lProperties->iAnisotropicFrictionScaling[2];
	shapeProps->m_do_anisotropic = lProperties->iHasAnisotropicFriction;
	
	shapeProps->m_do_fh     = lProperties->iDoFH; 
	shapeProps->m_do_rot_fh = lProperties->iDoRotFH;
	
	return shapeProps;
}

/*
static float my_boundbox_mesh(::Mesh *me, float *loc, float *size)
{
	MVert *mvert;
	BoundBox *bb;
	SimdPoint3 min, max;
	float mloc[3], msize[3];
	int a;
	
	if(me->bb==0) me->bb= (struct BoundBox *)MEM_callocN(sizeof(BoundBox), "boundbox");
	bb= me->bb;
	
	INIT_MINMAX(min, max);

	if (!loc) loc= mloc;
	if (!size) size= msize;
	
	mvert= me->mvert;
	for(a=0; a<me->totvert; a++, mvert++) {
		DO_MINMAX(mvert->co, min, max);
	}
		
	if(me->totvert) {
		loc[0]= (min[0]+max[0])/2.0;
		loc[1]= (min[1]+max[1])/2.0;
		loc[2]= (min[2]+max[2])/2.0;
		
		size[0]= (max[0]-min[0])/2.0;
		size[1]= (max[1]-min[1])/2.0;
		size[2]= (max[2]-min[2])/2.0;
	}
	else {
		loc[0]= loc[1]= loc[2]= 0.0;
		size[0]= size[1]= size[2]= 0.0;
	}
		
	bb->vec[0][0]=bb->vec[1][0]=bb->vec[2][0]=bb->vec[3][0]= loc[0]-size[0];
	bb->vec[4][0]=bb->vec[5][0]=bb->vec[6][0]=bb->vec[7][0]= loc[0]+size[0];
		
	bb->vec[0][1]=bb->vec[1][1]=bb->vec[4][1]=bb->vec[5][1]= loc[1]-size[1];
	bb->vec[2][1]=bb->vec[3][1]=bb->vec[6][1]=bb->vec[7][1]= loc[1]+size[1];

	bb->vec[0][2]=bb->vec[3][2]=bb->vec[4][2]=bb->vec[7][2]= loc[2]-size[2];
	bb->vec[1][2]=bb->vec[2][2]=bb->vec[5][2]=bb->vec[6][2]= loc[2]+size[2];

	float radius = 0;
	for (a=0, mvert = me->mvert; a < me->totvert; a++, mvert++)
	{
		float vert_radius = SimdVector3(mvert->co).length2();
		if (vert_radius > radius)
			radius = vert_radius;
	} 
	return sqrt(radius);
}
		



static void my_tex_space_mesh(::Mesh *me)
{
	KeyBlock *kb;
	float *fp, loc[3], size[3], min[3], max[3];
	int a;

	my_boundbox_mesh(me, loc, size);
	
	if(me->texflag & AUTOSPACE) {
		if(me->key) {
			kb= me->key->refkey;
			if (kb) {
	
				INIT_MINMAX(min, max);
		
				fp= (float *)kb->data;
				for(a=0; a<kb->totelem; a++, fp+=3) {	
					DO_MINMAX(fp, min, max);
				}
				if(kb->totelem) {
					loc[0]= (min[0]+max[0])/2.0; loc[1]= (min[1]+max[1])/2.0; loc[2]= (min[2]+max[2])/2.0;
					size[0]= (max[0]-min[0])/2.0; size[1]= (max[1]-min[1])/2.0; size[2]= (max[2]-min[2])/2.0;
	} 
	else {
					loc[0]= loc[1]= loc[2]= 0.0;
					size[0]= size[1]= size[2]= 0.0;
				}
				
			}
				}
	
		VECCOPY(me->loc, loc);
		VECCOPY(me->size, size);
		me->rot[0]= me->rot[1]= me->rot[2]= 0.0;
	
		if(me->size[0]==0.0) me->size[0]= 1.0;
		else if(me->size[0]>0.0 && me->size[0]<0.00001) me->size[0]= 0.00001;
		else if(me->size[0]<0.0 && me->size[0]> -0.00001) me->size[0]= -0.00001;
	
		if(me->size[1]==0.0) me->size[1]= 1.0;
		else if(me->size[1]>0.0 && me->size[1]<0.00001) me->size[1]= 0.00001;
		else if(me->size[1]<0.0 && me->size[1]> -0.00001) me->size[1]= -0.00001;
						
		if(me->size[2]==0.0) me->size[2]= 1.0;
		else if(me->size[2]>0.0 && me->size[2]<0.00001) me->size[2]= 0.00001;
		else if(me->size[2]<0.0 && me->size[2]> -0.00001) me->size[2]= -0.00001;
	}
	
}

static void my_get_local_bounds(::Object *ob, float *centre, float *size)
{
	BoundBox *bb= NULL;
	// uses boundbox, function used by Ketsji
	switch (ob->type)
	{
		case OB_MESH:
			bb= ( (::Mesh *)ob->data )->bb;
			if(bb==0) 
			{
				my_tex_space_mesh((struct ::Mesh *)ob->data);
				bb= ( (::Mesh *)ob->data )->bb;
			}
			break;
		case OB_CURVE:
		case OB_SURF:
		case OB_FONT:
			centre[0]= centre[1]= centre[2]= 0.0;
			size[0]  = size[1]=size[2]=0.0;
			break;
		case OB_MBALL:
			bb= ob->bb;
			break;
	}
	
	if(bb==NULL) 
	{
		centre[0]= centre[1]= centre[2]= 0.0;
		size[0] = size[1]=size[2]=1.0;
	}
	else 
	{
		size[0]= 0.5*fabs(bb->vec[0][0] - bb->vec[4][0]);
		size[1]= 0.5*fabs(bb->vec[0][1] - bb->vec[2][1]);
		size[2]= 0.5*fabs(bb->vec[0][2] - bb->vec[1][2]);
					
		centre[0]= 0.5*(bb->vec[0][0] + bb->vec[4][0]);
		centre[1]= 0.5*(bb->vec[0][1] + bb->vec[2][1]);
		centre[2]= 0.5*(bb->vec[0][2] + bb->vec[1][2]);
	}
}
*/
								
								
void AddToPhysicsEnv::CreatePhysicsObjectNew( grn::Object* gameobj
										    , grn::PhysicsProperties* lProperties
										    , PHY_IPhysicsEnvironment* aEnvironment )					
{

	PHY_ShapeProps* shapeprops = CreateShapePropsFromBlenderObject( lProperties );

	KX_ObjectProperties objprop;
	if ((objprop.m_isactor = lProperties->iActor))
	{
		objprop.m_dyna = lProperties->iDynamic;
		objprop.m_angular_rigidbody = lProperties->iRigidBody;
		objprop.m_ghost = lProperties->iGhost;
	} else {
		objprop.m_dyna = false;
		objprop.m_angular_rigidbody = false;
		objprop.m_ghost = false;
	}
	//mmm, for now, taks this for the size of the dynamicobject
	// Blender uses inertia for radius of dynamic object
	objprop.m_radius = lProperties->iRadius;
//	objprop.m_in_active_layer = (blenderobject->lay & activeLayerBitInfo) != 0;
	objprop.m_in_active_layer = true;
	objprop.m_dynamic_parent=NULL;
	objprop.m_isdeformable = lProperties->iDeformable;
	objprop.m_boundclass = objprop.m_dyna?KX_BOUNDSPHERE:KX_BOUNDMESH;
	
	KX_BoxBounds bb;
	objprop.m_boundobject.box.m_center[0] = lProperties->iLocalBoundsCenter.x();
	objprop.m_boundobject.box.m_center[1] = lProperties->iLocalBoundsCenter.y();
	objprop.m_boundobject.box.m_center[2] = lProperties->iLocalBoundsCenter.z();
	bb.m_extends[0] = lProperties->iLocalBoundsExtends.x();
	bb.m_extends[1] = lProperties->iLocalBoundsExtends.y();
	bb.m_extends[2] = lProperties->iLocalBoundsExtends.z();
	//my_get_local_bounds(blenderobject,objprop.m_boundobject.box.m_center,bb.m_extends);
	
	if (lProperties->iBoundType != PhysicsProperties::eBoundNone)
	{
		switch (lProperties->iBoundType)
		{
			case PhysicsProperties::eBoundBox:
				objprop.m_boundclass = KX_BOUNDBOX;
				//mmm, has to be divided by 2 to be proper extends
				objprop.m_boundobject.box.m_extends[0]=2.f*bb.m_extends[0];
				objprop.m_boundobject.box.m_extends[1]=2.f*bb.m_extends[1];
				objprop.m_boundobject.box.m_extends[2]=2.f*bb.m_extends[2];
				break;
			case PhysicsProperties::eBoundPolyt:
					objprop.m_boundclass = KX_BOUNDPOLYTOPE;
					break;
				// Object is not a mesh... fall through OB_BOUND_POLYH to 
				// OB_BOUND_SPHERE
			case PhysicsProperties::eBoundPolyh:
					objprop.m_boundclass = KX_BOUNDMESH;
					break;
				// Object is not a mesh... can't use polyheder. 
				// Fall through and become a sphere.
			case PhysicsProperties::eBoundSphere:
			{
				objprop.m_boundclass = KX_BOUNDSPHERE;
				objprop.m_boundobject.c.m_radius = SimdMax(bb.m_extends[0], SimdMax(bb.m_extends[1], bb.m_extends[2]));
				break;
			}
			case PhysicsProperties::eBoundCylinder:
			{
				objprop.m_boundclass = KX_BOUNDCYLINDER;
				objprop.m_boundobject.c.m_radius = SimdMax(bb.m_extends[0], bb.m_extends[1]);
				objprop.m_boundobject.c.m_height = 2.f*bb.m_extends[2];
				break;
			}
			case PhysicsProperties::eBoundCone:
			{
				objprop.m_boundclass = KX_BOUNDCONE;
				objprop.m_boundobject.c.m_radius = SimdMax(bb.m_extends[0], bb.m_extends[1]);
				objprop.m_boundobject.c.m_height = 2.f*bb.m_extends[2];
				break;
			}
		}
	}

/*
	// get Root Parent of blenderobject
	struct Object* parent= blenderobject->parent;
	while(parent && parent->parent) {
		parent= parent->parent;
	}

	if (parent && (parent->gameflag & OB_DYNAMIC)) {
		
		KX_GameObject *parentgameobject = converter->FindGameObject(parent);
		objprop.m_dynamic_parent = parentgameobject;

	}
*/
	objprop.m_concave = lProperties->iConcave;
	
	ConvertBulletObject(gameobj, aEnvironment, shapeprops, lProperties, &objprop);
}
 




/*
static CollisionShape* CreateBulletShapeFromMesh(RAS_MeshObject* meshobj, bool polytope)
{
	if (!meshobj)
		return 0;

	CollisionShape* collisionMeshShape = 0;
	ConvexHullShape* convexHullShape = 0;
	TriangleMeshShape* concaveShape = 0;

	TriangleMesh* collisionMeshData = 0;

	//see if there is any polygons, if not, bail out.

	int numUsedPolygons = 0;
	int numPoints = 0;
	SimdVector3* points = 0;

	CollisionShape** shapeptr = map_gamemesh_to_bulletshape[GEN_HashedPtr(meshobj)];

	// Mesh has already been converted: reuse
	if (shapeptr)
	{
		//return *shapeptr;
	}

	// Mesh has no polygons!
	int numpolys = meshobj->NumPolygons();
	if (!numpolys)
	{
		return NULL;
	}

	// Count the number of collision polygons and check they all come from the same 
	// vertex array
	int numvalidpolys = 0;
	int vtxarray = -1;
	RAS_IPolyMaterial *poly_material = NULL;
	bool reinstance = true;

	for (int p=0; p<numpolys; p++)
	{
		RAS_Polygon* poly = meshobj->GetPolygon(p);

		// only add polygons that have the collisionflag set
		if (poly->IsCollider())
		{
			// check polygon is from the same vertex array
			if (poly->GetVertexIndexBase().m_vtxarray != vtxarray)
			{
				if (vtxarray < 0)
					vtxarray = poly->GetVertexIndexBase().m_vtxarray;
				else
				{
					reinstance = false;
					vtxarray = -1;
				}
			}

			// check poly is from the same material
			if (poly->GetMaterial()->GetPolyMaterial() != poly_material)
			{
				if (poly_material)
				{
					reinstance = false;
					poly_material = NULL;
				}
				else
					poly_material = poly->GetMaterial()->GetPolyMaterial();
			}

			// count the number of collision polys
			numvalidpolys++;

			// We have one collision poly, and we can't reinstance, so we
			// might as well break here.
			if (!reinstance)
				break;
		}
	}

	// No collision polygons
	if (numvalidpolys < 1)
		return NULL;


	if (polytope)
	{
		convexHullShape = new ConvexHullShape(points,numPoints);
		collisionMeshShape = convexHullShape;
	} else
	{
		collisionMeshData = new TriangleMesh();
		concaveShape = new TriangleMeshShape(collisionMeshData);
		collisionMeshShape = concaveShape;

	}


	numvalidpolys = 0;

	for (int p2=0; p2<numpolys; p2++)
	{
		RAS_Polygon* poly = meshobj->GetPolygon(p2);

		// only add polygons that have the collisionflag set
		if (poly->IsCollider())
		{   
			//Bullet can raycast any shape, so
			if (polytope)
			{
				for (int i=0;i<poly->VertexCount();i++)
				{
					const float* vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[i],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 point(vtx[0],vtx[1],vtx[2]);
					convexHullShape->AddPoint(point);
				}
				if (poly->VertexCount())
					numvalidpolys++;

			} else
			{
				{
					const float* vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[2],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex0(vtx[0],vtx[1],vtx[2]);
					vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[1],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex1(vtx[0],vtx[1],vtx[2]);
					vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[0],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex2(vtx[0],vtx[1],vtx[2]);
					collisionMeshData->AddTriangle(vertex0,vertex1,vertex2);
					numvalidpolys++;
				}
				if (poly->VertexCount() == 4)
				{
					const float* vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[3],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex0(vtx[0],vtx[1],vtx[2]);
					vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[2],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex1(vtx[0],vtx[1],vtx[2]);
					vtx = meshobj->GetVertex(poly->GetVertexIndexBase().m_vtxarray, 
						poly->GetVertexIndexBase().m_indexarray[0],
						poly->GetMaterial()->GetPolyMaterial())->getLocalXYZ();
					SimdPoint3 vertex2(vtx[0],vtx[1],vtx[2]);
					collisionMeshData->AddTriangle(vertex0,vertex1,vertex2);
					numvalidpolys++;
				}

			}		
		}
	}



	if (numvalidpolys > 0)
	{
		//map_gamemesh_to_bulletshape.insert(GEN_HashedPtr(meshobj),collisionMeshShape);
		return collisionMeshShape;
	}

	delete collisionMeshShape;
	return NULL;

}
*/

class MotionState : public PHY_IMotionState
{
public:
	MotionState( grn::Object* aObject )
	: iObject(aObject)
	{
	};
	
protected:

	Pointer<grn::Object> iObject;

	SimdPoint3 iLastWorldPosSet;
	SimdQuaternion iLastWorldRotSet;
	
// From PHY_IMotionState
public:
		virtual void getWorldPosition(float& posX,float& posY,float& posZ)
		{
			SimdTransform lWT;
			iObject->GetWorldTransform(lWT);
			posX = lWT.getOrigin().x();
			posY = lWT.getOrigin().y();
			posZ = lWT.getOrigin().z();
		}
		
		
		// What is this?
		virtual void getWorldScaling(float& scaleX,float& scaleY,float& scaleZ)
		{
			/*
			scaleX = iTransform->iTransform.getBasis()[0].length();
			scaleY = iTransform->iTransform.getBasis()[1].length();
			scaleZ = iTransform->iTransform.getBasis()[2].length();
			*/
			scaleX = 1.0f;
			scaleY = 1.0f;
			scaleZ = 1.0f;
		}
		
		
		virtual void getWorldOrientation(float& quatIma0,float& quatIma1,float& quatIma2,float& quatReal)
		{
			SimdTransform lWT;
			iObject->GetWorldTransform(lWT);
			SimdQuaternion q = lWT.getRotation();
			quatIma0 = q.x();
			quatIma1 = q.y();
			quatIma2 = q.z();
			quatReal = q[3];
		}
		
		
		virtual void setWorldPosition(float posX,float posY,float posZ)
		{
			iLastWorldPosSet = SimdPoint3(posX, posY, posZ);
		}
		
		virtual	void setWorldOrientation(float quatIma0,float quatIma1,float quatIma2,float quatReal)
		{
			iLastWorldRotSet = SimdQuaternion(quatIma0, quatIma1, quatIma2, quatReal);
		}
		
		// Called by the physics controller, after setting position and orientation.
		virtual	void calculateWorldTransformations()
		{
			iObject->SetWorldTransform( SimdTransform(iLastWorldRotSet,iLastWorldPosSet) );
		}
	
};


void ConvertBulletObject( grn::Object* gameobj
						, PHY_IPhysicsEnvironment* aEnvironment
						, PHY_ShapeProps* shapeprops
						, grn::PhysicsProperties*	lProperties
						, KX_ObjectProperties*	objprop )
{
	if (!objprop->m_isactor	) return;


	CcdPhysicsEnvironment* env = (CcdPhysicsEnvironment*)aEnvironment;
	assert(env);
	

	bool dyna = false;
	CcdConstructionInfo ci;
	//gameobj->iPhysicsTransform = new grn::Transform();
	PHY_IMotionState* motionstate = new MotionState(gameobj);

	ci.m_MotionState = motionstate;
	ci.m_gravity = SimdVector3(0,0,0);
	ci.m_localInertiaTensor =SimdVector3(0,0,0);
	ci.m_mass = objprop->m_dyna ? shapeprops->m_mass : 0.f;
	
	ci.m_localInertiaTensor = SimdVector3(ci.m_mass/3.f,ci.m_mass/3.f,ci.m_mass/3.f);
	
	SimdTransform trans;
	trans.setIdentity();
	
	CollisionShape* bm = 0;

	switch (objprop->m_boundclass)
	{
	case KX_BOUNDSPHERE:
		{
			float radius = objprop->m_radius;
			SimdVector3 inertiaHalfExtents (
				radius,
				radius,
				radius);
			
			//blender doesn't support multisphere, but for testing:

			//bm = new MultiSphereShape(inertiaHalfExtents,,&trans.getOrigin(),&radius,1);
			bm = new SphereShape(objprop->m_radius);
			bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);
			break;
		};
	case KX_BOUNDBOX:
		{
			SimdVector3 halfExtents (
				objprop->m_boundobject.box.m_extends[0],
			objprop->m_boundobject.box.m_extends[1],
			objprop->m_boundobject.box.m_extends[2]);

			halfExtents /= 2.f;

			SimdVector3 he (halfExtents[0]-CONVEX_DISTANCE_MARGIN ,halfExtents[1]-CONVEX_DISTANCE_MARGIN ,halfExtents[2]-CONVEX_DISTANCE_MARGIN );
			he = he.absolute();


			bm = new BoxShape(he);
			bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);

			break;
		};
	case KX_BOUNDCYLINDER:
		{
			SimdVector3 halfExtents (
				objprop->m_boundobject.c.m_radius,
				objprop->m_boundobject.c.m_radius,
				objprop->m_boundobject.c.m_height * 0.5f
			);
			bm = new CylinderShapeZ(halfExtents);
			bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);

			break;
		}

		case KX_BOUNDCONE:
		{
				SimdVector3 halfExtents (objprop->m_boundobject.box.m_extends[0],
				objprop->m_boundobject.box.m_extends[1],
				objprop->m_boundobject.box.m_extends[2]);


				halfExtents /= 2.f;
				SimdVector3&	he  = halfExtents;
				SimdTransform& tr = trans;

				bm = new ConeShape(objprop->m_boundobject.c.m_radius,objprop->m_boundobject.c.m_height);
				bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);



			break;
		}
		/*
		case KX_BOUNDPOLYTOPE:
			{
				bm = CreateBulletShapeFromMesh(meshobj,true);
				if (bm)
				{
					bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);
				}
				break;
			}
		case KX_BOUNDMESH:
			{
				if (!ci.m_mass)
				{				
					bm = CreateBulletShapeFromMesh(meshobj,false);
					ci.m_localInertiaTensor.setValue(0.f,0.f,0.f);
					//no moving concave meshes, so don't bother calculating inertia
					//bm->CalculateLocalInertia(ci.m_mass,ci.m_localInertiaTensor);
				}

				break;
			}
		*/
	default:
		//interpret the shape as a concave triangle-mesh
		{
			/*
			if (meshobj)
			{
			//	assert(0);

				meshobj->ScheduleCollisionPolygons();

				KX_DeformableMesh* gfxmesh = new KX_DeformableMesh(meshobj);
				gfxmesh->sendFixedMapping();
				//trianglemesh
				bm = new TriangleMeshInterface(gfxmesh,trans);
			}
				*/
		}
	}


//	ci.m_localInertiaTensor.setValue(0.1f,0.1f,0.1f);

	if (!bm)
		return;
		

	ci.m_collisionShape = bm;
	ci.m_broadphaseHandle = 0;
	ci.m_friction = lProperties->iFriction;
	ci.m_restitution = lProperties->iRestitution;


	ci.m_linearDamping = shapeprops->m_lin_drag;
	ci.m_angularDamping = shapeprops->m_ang_drag;

	CcdPhysicsController* physicscontroller = new CcdPhysicsController(ci);
	env->addCcdPhysicsController( physicscontroller);

	
	gameobj->iPhysicsController = physicscontroller;
	physicscontroller->setNewClientInfo(gameobj);						
//	gameobj->GetSGNode()->AddSGController(physicscontroller);

/*
	bool isActor = objprop->m_isactor;
	STR_String materialname;
	if (meshobj)
		materialname = meshobj->GetMaterialName(0);

	const char* matname = materialname.ReadPtr();
*/

	//physicscontroller->SetObject(gameobj->GetSGNode());
				

}
