
#include "CollisionQuery.h"

#include "engine/Engine.h"
#include "GironaLog.h"
#include "Pipeline.h"

// bullet
#include "NarrowPhaseCollision/PersistentManifold.h"

using namespace grn;

IMPLEMENT_MODULE(CollisionQuery);
				

CollisionQuery::CollisionQuery( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,70);
	iX = 500;
	riSetA = new Module::tInputPad<Object>( "ObjectSet1", this );
	riSetB = new Module::tInputPad<Object>( "ObjectSet2", this );
	riCollisions = new Module::tInputPad<CollisionList>( "CollisionList", this );
	roResult = new Module::tOutputPad<Boolean>( "Result", this );
}


void CollisionQuery::Run()
{
	while ( AllInputsHaveData() )
	{	
		bool bRes = false;
		
		// Get collision list
		Pointer<CollisionList> lCol = riCollisions->getData(true);
		
		// Get all sets
		std::vector< Pointer<Object> > lSetA;
		while (riSetA->hasData()) lSetA.push_back(riSetA->getData(true));
		
		std::vector< Pointer<Object> > lSetB;
		while (riSetB->hasData()) lSetB.push_back(riSetB->getData(true));

		for ( std::vector< Pointer<Object> >::iterator it = lSetA.begin()
			; (it != lSetA.end()) && (!bRes)
			; it++ )
			if ((*it)->iPhysicsController)
			{
				const void* lRB1 = (*it)->iPhysicsController->GetRigidBody();
				
				for ( std::vector< Pointer<Object> >::iterator it2 = lSetB.begin()
					; (it2 != lSetB.end()) && (!bRes)
					; it2++ )
					if ((*it2)->iPhysicsController)
					{
						const void* lRB2 = (*it2)->iPhysicsController->GetRigidBody();
						
						// Look for pair in collision list
						for ( std::vector<CollisionList::CollisionPair>::iterator it3 = lCol->iPairs.begin()
							; (it3 != lCol->iPairs.end()) && (!bRes)
							; it3++)
							bRes |= ( (lRB1==it3->first) && (lRB2==it3->second) )
								 || ( (lRB1==it3->second) && (lRB2==it3->first) );
					}
			}
					

		if (bRes) 
			log(" EEEEEEEEEEmited collisions ");
		

		Pointer<Boolean> lRes = new Boolean();
		lRes->iValue = bRes;

		rPipeline->Emit(roResult, lRes);
	}
}


