
#include "UpdatePhysicsEnv.h"

#include "engine/Engine.h"
#include "GironaLog.h"
#include "Pipeline.h"

// bullet
#include "NarrowPhaseCollision/PersistentManifold.h"

using namespace grn;

IMPLEMENT_MODULE(UpdatePhysicsEnv);
				

UpdatePhysicsEnv::UpdatePhysicsEnv( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,70);
	iX = 500;
	riEnv = new Module::tInputPad<PhysicsEnvironment>( "Environment", this );
	roCollisions = new Module::tOutputPad<CollisionList>( "CollisionList", this );
	rEngine = NULL;
}


void UpdatePhysicsEnv::Link( Engine* aEngine )
{
	rEngine = aEngine;
}


void UpdatePhysicsEnv::Run()
{
	assert(rEngine);

	while ( AllInputsHaveData() )
	{	
		Pointer<PhysicsEnvironment> lEnv = riEnv->getData(true);
	
		Pointer<CollisionList> lCol = new CollisionList();
		
		// Advance the physics	
		
		//#ifndef GIRONA_PSP
		lEnv->Update( rEngine->GetCurrentTime(), rEngine->GetDeltaTime() );
		//#endif
		
		// Check collisions
		for ( int i=0; i<lEnv->iPhysicsEnvironment->GetNumManifolds(); i++ )
		{
			const PersistentManifold* lPM = lEnv->iPhysicsEnvironment->GetManifold(i);
			if (lPM->GetNumContacts())
			{
				lCol->iPairs.push_back( std::pair<const void*, const void*>(lPM->GetBody0(), lPM->GetBody1()) );
			}
		}
		
		rPipeline->Emit(roCollisions, lCol);
	}
	
}


