
#include "LinkHierarchy.h"

#include "Pipeline.h"


using namespace grn;


IMPLEMENT_MODULE(LinkHierarchy);


LinkHierarchy::LinkHierarchy( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,60);
	iX = 300;
	riObjects = new Module::tInputPad<grn::Object>( "InputObjects", this );
	roObjects = new Module::tOutputPad<grn::Object>( "LinkedObjects", this );
}


void LinkHierarchy::Init()
{
	iObjectsProcessed.clear();
	iObjectsPending.clear();
}


void LinkHierarchy::Run()
{
	if ( riObjects->hasData() )
	{
		std::cout<<" Linker "<<std::endl;

		// Get all the input objects
		while (riObjects->hasData()) 
		{
			Pointer<Object> lObject = riObjects->getData(true);
			iObjectsPending.push_back( lObject );
			iObjectsProcessed.push_back( lObject );
		}
		
		// For each object not yet linked, try to find the parent.
		for ( ObjectList::iterator it = iObjectsPending.begin()
			; it != iObjectsPending.end()
			; )
		{
			std::string lOrphanParentName = (*it)->iParentName->iValue;

			bool lFound =  (lOrphanParentName == kNullObjectName);

			// Look into all the processed objects
			for ( ObjectList::iterator itt = iObjectsProcessed.begin()
				; (itt != iObjectsProcessed.end()) && (!lFound)
				; itt++ )
				{
					std::string lCandidateParent = (*itt)->iName->iValue;
					std::cout<<"    Comparing "<<lCandidateParent<<" with "<<lOrphanParentName<<std::endl;
					if ( lCandidateParent == lOrphanParentName )
					{					
						lFound = true;
						
						// Link to parent
						(*it)->Relink(*itt);
						
					}
				}				
			if (!lFound) it++;
			else
			{
				// ready to be emmited
				rPipeline->Emit(roObjects,*it);
				
				// remove from pending list
				it = iObjectsPending.erase(it);
			}
		}
		
	}
}
