
/*
#include <GL/glut.h>
#include <cstdio>
#include <cstdlib>

#ifdef _WIN32
#	include <windows.h>
#else
#	include <sys/time.h>
#endif

#include "PlayerEngine.h"

#include "girona/RootPipeline.h"
#include "girona/Serialise.h"

#include <pspmoduleinfo.h>


PSP_MODULE_INFO("GironaPlayer", 0, 1, 1);



static int g_Width = 512;
static int g_Height = 512;


#ifdef _WIN32
static DWORD last_idle_time;
#else
static struct timeval last_idle_time;
#endif



static PlayerEngine* gEngine(0);
static grn::RootPipeline* gRoot(0);
static bool gExit(false);


void display(void)
{
	float lDeltaTime;
	
#ifdef _WIN32
	DWORD time_now;
	time_now = GetTickCount();
	lDeltaTime = (float) (time_now - last_idle_time) / 1000.0;
#else
	// Figure out time elapsed since last call to idle function
	struct timeval time_now;
	gettimeofday(&time_now, NULL);
	lDeltaTime = (float)(time_now.tv_sec  - last_idle_time.tv_sec) + 1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
#endif

	
//	double lCurrentTime = PIL_check_seconds_timer();
	double lCurrentTime = 0.0;
	
	// Tell the engine the delta time
	gEngine->SetDeltaTime(lDeltaTime);
	gEngine->SetCurrentTime(lCurrentTime);

	// Run a frame
	gEngine->RunFrame();

	// Save time_now for next time
	last_idle_time = time_now;
}


void Keyboard(unsigned char key, int , int )
{
#ifndef GIRONA_PSP	
	if (key==27) exit(0);
#else
	if (key=='x') exit(0);
#endif
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], true);
}

void KeyboardUp(unsigned char key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], false);
}


void SpecialKeyboard(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], true);
}

void SpecialKeyboardUp(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], false);
}


void loop()
{
	// Force redraw
	glutPostRedisplay();
	
}

 
int main(int argc, char** argv)
{
	// GLUT Window Initialization:
	glutInit (&argc, argv);
	glutInitWindowSize (g_Width, g_Height);
	glutInitDisplayMode ( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow ("Blender-Girona Player");

	// Initialize the game logic
	gRoot = new grn::RootPipeline();
	gRoot->Load("unnamed.girona");
	
	// Create engine class
	gEngine = new PlayerEngine( gRoot );
	
	// Load game data
	gEngine->LoadData( "unnamed.data" );

	// Set up the game render pipeline
	gEngine->SetScreenArea( 0, 0, g_Width, g_Height );
	gEngine->Start();
	
	gExit = false;

	// Register callbacks:
	glutDisplayFunc (display);
	glutReshapeFunc (reshape);
	glutKeyboardFunc (Keyboard);
	glutKeyboardUpFunc (KeyboardUp);
	glutSpecialFunc (SpecialKeyboard);
	glutSpecialUpFunc (SpecialKeyboardUp);
	//glutMouseFunc (MouseButton);
	//glutMotionFunc (MouseMotion);
	glutIdleFunc (loop);


	// Get the initial time
	gettimeofday (&last_idle_time, NULL);

	// Turn the flow of control over to GLUT
	glutMainLoop ();
		
	gRoot->ClearData();
		
	delete gEngine;

	delete gRoot;
		
	return 1;
	
}

*/


#include <stdlib.h>
#include <GL/glut.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>


#include "PlayerEngine.h"
#include "girona/engine/GL2/RendererGL2.h"
#include "girona/RootPipeline.h"
#include "girona/Serialise.h"

static PlayerEngine* gEngine(0);
static grn::RootPipeline* gRoot(0);
static bool gExit(false);
static struct timeval last_idle_time;


#define DATA_PATH_PREFIX std::string("")



static void display(void)
{
	float lDeltaTime;
	
	// Figure out time elapsed since last call to idle function
	struct timeval time_now;
	gettimeofday(&time_now, NULL);
	lDeltaTime = (float)(time_now.tv_sec  - last_idle_time.tv_sec) + 1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
	
	double lCurrentTime = 0.0;
	
	// Tell the engine the delta time
	gEngine->SetDeltaTime(lDeltaTime);
	gEngine->SetCurrentTime(lCurrentTime);

	// Run a frame
	gEngine->RunFrame();

	// Save time_now for next time
	last_idle_time = time_now;

	// Order another frame
	glutPostRedisplay();

}


static void Keyboard(unsigned char key, int , int )
{
#ifndef GIRONA_PSP	
	if (key==27) exit(0);
#else
	if (key=='x') exit(0);
#endif
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], true);
}

static void KeyboardUp(unsigned char key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], false);
}


static void SpecialKeyboard(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], true);
}

static void SpecialKeyboardUp(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], false);
}


static void reshape (int w, int h)
{
	grn::log("Reshaping",h);
	
	// Set up the game render pipeline
	gEngine->SetScreenArea( 0, 0, w, h );
	gEngine->Start();
}


int main(int argc, char* argv[])
{
	glutInit(&argc, argv);
	glutInitWindowSize ( 512, 512 );
	glutInitDisplayMode ( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow( __FILE__ );
	
	glutKeyboardFunc(Keyboard);
	glutKeyboardUpFunc(KeyboardUp);
	glutSpecialFunc (SpecialKeyboard);
	glutSpecialUpFunc (SpecialKeyboardUp);
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);

	grn::log("Before initialising Girona\n");

	// Initialize the game logic
	gRoot = new grn::RootPipeline();

	grn::log("Before loading logic\n");
	
	gRoot->Load(DATA_PATH_PREFIX+"unnamed.girona");
	
	grn::log("Pipelines : ",(int)gRoot->iModules.size());

	// Create engine class
	grn::log("Before creating engine\n");
	grn::Renderer* lRenderer = new grn::RendererGL2();	
	gEngine = new PlayerEngine( gRoot, lRenderer );

	// Load game data
	grn::log("Befor loading data\n");

	gEngine->LoadData(DATA_PATH_PREFIX+"unnamed.data" );

	gExit = false;

	// Get the initial time
	gettimeofday (&last_idle_time, NULL);

	grn::log("After initialising Girona\n");

	glutMainLoop();
	
	delete gEngine;
	delete lRenderer;

	return 0;
}

