

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspgu.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

PSP_MODULE_INFO("GironaPlayer", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER);


#include "PlayerEngine.h"
#include "girona/engine/RendererGU.h"
#include "girona/RootPipeline.h"
#include "girona/Serialise.h"

static PlayerEngine* gEngine(0);
static grn::RootPipeline* gRoot(0);
static struct timeval last_idle_time;


#define DATA_PATH_PREFIX std::string("ms0:/")


int SetupCallbacks();

#include <stdarg.h>

extern "C"
{
	
void __psp_log (const char *fmt, ...)
{
	char buf [1024];
	va_list ap;
	int len;
	int log_fd;

	log_fd = sceIoOpen("ms0:/log.txt", PSP_O_CREAT | PSP_O_APPEND | PSP_O_WRONLY, 0644);

	if (log_fd >= 0) {
		va_start(ap, fmt);
		len = vsnprintf(buf, sizeof(buf), fmt, ap);
		va_end(ap);
		sceIoWrite(log_fd, buf, len);
		sceIoClose(log_fd);
	}
}

}


int main(int argc, char* argv[])
{
	pspDebugScreenInit();

	SetupCallbacks();

	sceKernelDcacheWritebackAll();

	// setup GU

	sceGuInit();

	__psp_log("Loading \n");

//////////////////////////////////////

	// Initialize the game logic
	gRoot = new grn::RootPipeline();

	gRoot->Load(DATA_PATH_PREFIX+"unnamed.girona");	
	grn::log("Pipelines : ",(int)gRoot->iModules.size());

	// Create engine class
	grn::Renderer* lRenderer = new grn::RendererGU();	
	gEngine = new PlayerEngine( gRoot, lRenderer );

	// Load game data
	gEngine->LoadData(DATA_PATH_PREFIX+"unnamed.data" );

	// Set up the game render pipeline
	gEngine->SetScreenArea( 0, 0, 480, 272 );
	gEngine->Start();

	// Get the initial time
	gettimeofday (&last_idle_time, NULL);

/////////////////////////////////////


	// run sample
	__psp_log("Starting \n");

	float lCurrentTime = 0.0;
	
	for(;;)
	{

		float lDeltaTime;
		
		// Figure out time elapsed since last call to idle function
		struct timeval time_now;
		gettimeofday(&time_now, NULL);
		lDeltaTime = (float)(time_now.tv_sec  - last_idle_time.tv_sec) + 1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
		
		//__psp_log("Delta %f \n",lDeltaTime);
		
		lCurrentTime += lDeltaTime;
		
		// Tell the engine the delta time
		gEngine->SetDeltaTime(lDeltaTime);
		gEngine->SetCurrentTime(lCurrentTime);
	
		// Run a frame
		gEngine->RunFrame();
		
		gEngine->ProcessEvents();
	
		// Save time_now for next time
		last_idle_time = time_now;
	}

	sceGuTerm();

	sceKernelExitGame();
	return 0;
}

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}



