

#include "BlenderDraw.h"
#include "GironaLog.h"

#include "Pipeline.h"


extern "C" {

#include "BIF_gl.h"
#include "BIF_resources.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "BIF_butspace.h"
#include "BIF_screen.h"
#include "BIF_keyval.h"
#include "BIF_mywindow.h"
#include "PIL_time.h"

#include "DNA_object_types.h"
#include "DNA_screen_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"

#include "butspace.h"
#include "interface.h"
#include "mydevice.h"

}


#define LINK_SIZE 19

using namespace grn;


void module_update_links(void *, void *)
{
}
	
void module_select(void *selpropv, void *)
{
	Module* lThis = (Module*) selpropv;
	lThis->Select();
	allqueue(REDRAWBUTSLOGIC, curarea->win);	
}
	


void BlenderDrawVisitor::Visit( Module* aM )
{
	assert(aM->GetParent());
	
	if (!iSelected) uiBlockSetCol(iBlock, TH_AUTO);
	else uiBlockSetCol(iBlock, TH_REDALERT);
	
	uiBut* iBase = uiDefBut( iBlock, BUT, 1, const_cast<char*>(aM->GetModuleUserName().c_str()),		
					  aM->iX, aM->iY, aM->iSizeX, aM->iSizeY,
					  NULL, 0.0, 100.0, 100, 0,
					  "");
	uiButSetPressFunc(iBase, module_select, aM, NULL);

	int lSpan;
	if (aM->iInputs.size())
	{
		lSpan = aM->iSizeY/aM->iInputs.size();
		for ( unsigned int i=0; i<aM->iInputs.size(); i++ )
		{
			uiBut* iLink = uiDefIconBut( iBlock, INLINK, 0, ICON_INLINK
						  , aM->iX-LINK_SIZE, aM->iY+i*lSpan+lSpan/2-LINK_SIZE/2
						  , LINK_SIZE, LINK_SIZE
						  , aM->iInputs[i], aM->GetParent()->GetDataTypeID(aM->iInputs[i]->iType)
						  , 0, 0, 0
						  , const_cast<char*>(aM->iInputs[i]->iName.c_str()) );
			uiButSetFunc( iLink, module_update_links, aM, iLink );
		}
	}
	
	if (aM->iOutputs.size())
	{
		lSpan = aM->iSizeY/aM->iOutputs.size();
		for ( unsigned int i=0; i<aM->iOutputs.size(); i++ )
		{						  
			uiBut* iLink = uiDefIconBut( iBlock, LINK, 0, ICON_LINK
										, aM->iX+aM->iSizeX, aM->iY+i*lSpan+lSpan/2-LINK_SIZE/2
										, LINK_SIZE, LINK_SIZE
								  		, NULL, 0, 0, 0, 0
								  		, const_cast<char*>(aM->iOutputs[i]->iName.c_str()) );

			uiSetButLink( iLink, NULL
						, (void ***)&(aM->iOutputs[i]->iLinks), &aM->iOutputs[i]->iNumLinks
						, aM->GetParent()->GetDataTypeID(aM->iOutputs[i]->iType), aM->GetParent()->GetDataTypeID(aM->iOutputs[i]->iType) );
			uiButSetFunc( iLink, module_update_links, aM, iLink );

		}		
	}					  
}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


static void if_select_datatype(void *aModule, void *)
{
	If* lModule = (If*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( If* aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to select%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-20, aM->iSizeX, 20
							, &aM->iImportedType, 0.0, 0.0, 0, 0
							, "Selects the data type to select from a sequence.");
	uiButSetFunc( iBase, if_select_datatype, aM, NULL);
}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


static void endofdata_select_datatype(void *aModule, void *)
{
	EndOfData* lModule = (EndOfData*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( EndOfData *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to end%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-20, aM->iSizeX, 20
							, &aM->iType, 0.0, 0.0, 0, 0
							, "Selects the data type to import from outside this pipeline.");
	uiButSetFunc( iBase, endofdata_select_datatype, aM, NULL);
	
}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////

static void export_select_datatype(void *aModule, void *)
{
	Export* lModule = (Export*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( Export* aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	uiDefBut( iBlock, TEX, 0, "Name: "
			, aM->iX, aM->iY-20, aM->iSizeX, 20
			, aM->iName, 0.0, 79.0, 0, 0
			, "Displays the name of the link when exposed to parent pipeline. Click to change."
			);

	std::stringstream lMenu;
	lMenu<<"Type to export%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, 1, const_cast<char*>( lMenu.str().c_str() )
				, aM->iX, aM->iY-40, aM->iSizeX, 20
				, &(aM->iExportedType), 0.0, 0.0, 0, 0
				, "Selects the data type to export to parent pipelines.");
	uiButSetFunc( iBase, export_select_datatype, aM, NULL);
	
}



/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


static void import_select_datatype(void *aModule, void *)
{
	Import* lModule = (Import*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( Import *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	uiDefBut( iBlock, TEX, 0, "Name: "
			, aM->iX, aM->iY-20, aM->iSizeX, 20
			, aM->iName, 0.0, 79.0, 0, 0
			, "Displays the name of the link when exposed to parent pipeline. Click to change."
			);

	std::stringstream lMenu;
	lMenu<<"Type to import%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-40, aM->iSizeX, 20
							, &aM->iImportedType, 0.0, 0.0, 0, 0
							, "Selects the data type to import from outside this pipeline.");
	uiButSetFunc( iBase, import_select_datatype, aM, NULL);

	char* lStr = "Accumulate all data %t|False %x0| True%x1";
	iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, lStr
						, aM->iX, aM->iY-60, aM->iSizeX, 20
						, &aM->iAccumulate, 0.0, 0.0, 0, 0
						, "Accumulate. If activated, imported data will persist .");
	
}

/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////

static void markgroup_select_datatype(void *aModule, void *)
{
	MarkGroup* lModule = (MarkGroup*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( MarkGroup *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to import%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-20, aM->iSizeX, 20
							, &aM->iImportedType, 0.0, 0.0, 0, 0
							, "Selects the data type to import from outside this pipeline.");
	uiButSetFunc( iBase, markgroup_select_datatype, aM, NULL);

}



/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////

static void replicate_select_datatype(void *aModule, void *)
{
	Replicate* lModule = (Replicate*)aModule;
	lModule->UpdateSelectedTypes();
}


void BlenderDrawVisitor::Visit( Replicate* aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to select%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-20, aM->iSizeX, 20
							, &aM->iSourceType, 0.0, 0.0, 0, 0
							, "Selects the data type to replicate.");
	uiButSetFunc( iBase, replicate_select_datatype, aM, NULL);

	iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-40, aM->iSizeX, 20
							, &aM->iTimesType, 0.0, 0.0, 0, 0
							, "Selects the data type to count.");
	uiButSetFunc( iBase, replicate_select_datatype, aM, NULL);

}



/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


static void sequenceselect_select_datatype(void *aModule, void *)
{
	SequenceSelect* lModule = (SequenceSelect*)aModule;
	lModule->UpdateSelectedType();
}


void BlenderDrawVisitor::Visit( SequenceSelect *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to select%t";

	Pipeline::TypeMap& rTypeMap = aM->GetParent()->GetTypeMap();
	for ( Pipeline::TypeMap::iterator it = rTypeMap.begin()
		; it != rTypeMap.end()
		; it++ )
		lMenu<<"|"<<it->first<<" %x"<<it->second;
	
	uiBut* iBase = uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
							, aM->iX, aM->iY-20, aM->iSizeX, 20
							, &aM->iImportedType, 0.0, 0.0, 0, 0
							, "Selects the data type to select from a sequence.");
	uiButSetFunc( iBase, sequenceselect_select_datatype, aM, NULL);

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( SelectBlenderScene *aM )
{
	
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	uiDefBut( iBlock, TEX, 0, "Name: "
			, aM->iX, aM->iY-20, aM->iSizeX, 20
			, aM->iName, 0.0, 79.0, 0, 0
			, "Displays the name of the scene to select. Click to change."
			);

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////



void BlenderDrawVisitor::Visit( BlenderObjectFilter *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str1= "Object Types%t|Mesh %x0|Armature %x1|Camera %x2|SpotLight %x3";
	uiDefButS (iBlock, MENU, B_MATPRV_DRAW, str1
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iType, 0.0, 0.0, 0, 0
				, "Selects the object type that will pass through the filter.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(KeyEvent* aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	std::stringstream lMenu;
	lMenu<<"Type to select%t";
	for (int i=0;i<Engine::_eNumKeys; i++)
		lMenu<<"|"<<Engine::sKeyNames[i]<<"%x"<<i;

	uiDefButS (iBlock, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iKey, 0.0, 0.0, 0, 0
				, "Select key that triggers the event.");

	char *str2= "Action%t|On press %x0|On release %x1|While pressed %x2";
	uiDefButS (iBlock, MENU, B_MATPRV_DRAW, str2
				, aM->iX, aM->iY-40, aM->iSizeX, 20
				, &aM->iAction, 0.0, 0.0, 0, 0
				, "Select key action that triggers the event.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(PhysicsModifier *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str2= "Physics property%t|Linear velocity %x0|Angular velocity %x1|Apply force %x2|Apply torque %x3|Set orientation %x4";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, (short)(aM->iSizeX*0.75), 20
				, &aM->iProperty, 0.0, 0.0, 0, 0
				, "Select the physic property to alter.");
	uiDefButS (iBlock, TOGR, 0, "Local"
				, aM->iX+(short)(aM->iSizeX*0.75), aM->iY-20, (short)(aM->iSizeX*0.25), 20
				, &aM->iLocal, 0.0, 0.0, 0, 0
				, "Use object local coordinate system or the world's.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(SelectObject *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str2= "Selection criteria%t|by name %x0|by propery %x1";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iCriteria, 0.0, 0.0, 0, 0
				, "Set the cryterion to use to select the objects.");

	uiDefBut( iBlock, TEX, 0, ""
			, aM->iX, aM->iY-40, aM->iSizeX, 20
			, aM->iName, 0.0, 79.0, 0, 0
			, "Displays the name of the object to select. Click to change."
			);

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(AddReal *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str2= "Operation%t|+%x0|-%x1|*%x2|/%x3";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iOperation, 0.0, 0.0, 0, 0
				, "Set the cryterion to use to select the objects.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(CompareReal *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str2= "Operation%t|=%x0|<%x1|>%x2|!=%x3";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iOperation, 0.0, 0.0, 0, 0
				, "Set the comparison operation between the reals.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit(OperateBoolean *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	char *str2= "Operation%t|AND%x0|OR%x1|XOR%x2";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->iOperation, 0.0, 0.0, 0, 0
				, "Set the boolean operation to perform.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( ConstantInteger *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	uiDefButI (iBlock, NUM, 0, ""
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->i, -1000.0, 1000.0, 10, 0
				, "");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( ConstantReal *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	uiDefButF (iBlock, NUM|FLO, 0, ""
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->i, -1000.0, 1000.0, 10, 0
				, "");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( ConstantVector *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	

	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX, aM->iY-20, (short)(aM->iSizeX*0.3333), 20
				, &aM->ivX, -1000.0, 1000.0, 10, 0
				, "");
	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX+(short)(aM->iSizeX*0.3333), aM->iY-20, (short)(aM->iSizeX*0.3333), 20
				, &aM->ivY, -1000.0, 1000.0, 10, 0
				, "");
	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX+(short)(aM->iSizeX*0.6666), aM->iY-20, (short)(aM->iSizeX*0.3334), 20
				, &aM->ivZ, -1000.0, 1000.0, 10, 0
				, "");


}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( EnablePipeline *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	
	uiDefBut( iBlock, TEX, 0, "Pipeline:"
			, aM->iX, aM->iY-20, aM->iSizeX, 20
			, aM->iName, 0.0, 79.0, 0, 0
			, "Displays the name of the pipeline to enable or disable. Click to change."
			);

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( VariableVector *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);
	

	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX, aM->iY-20, (short)(aM->iSizeX*0.3333), 20
				, &aM->ivXinitial, -1000.0, 1000.0, 10, 0
				, "");
	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX+(short)(aM->iSizeX*0.3333), aM->iY-20, (short)(aM->iSizeX*0.3333), 20
				, &aM->ivYinitial, -1000.0, 1000.0, 10, 0
				, "");
	uiDefButF (iBlock, NUM, 0, ""
				, aM->iX+(short)(aM->iSizeX*0.6666), aM->iY-20, (short)(aM->iSizeX*0.3334), 20
				, &aM->ivZinitial, -1000.0, 1000.0, 10, 0
				, "");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( VariableInteger *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	uiDefButI (iBlock, NUM, 0, ""
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->initial, -1000.0, 1000.0, 10, 0
				, "");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( VariableBoolean *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	char *str2= "Operation%t|False%x0|True%x1";
	uiDefButS (iBlock, MENU, 0, str2
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->initial, 0.0, 0.0, 0, 0
				, "Set the initial boolean value.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( VariableReal *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	uiDefButF (iBlock, NUM|FLO, 0, ""
				, aM->iX, aM->iY-20, aM->iSizeX, 20
				, &aM->initial, -1000.0, 1000.0, 10, 0
				, "");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


void BlenderDrawVisitor::Visit( ConstantBoolean *aM )
{
	Visit( (Module*)aM );
	
	// Add GUI to control data sets
	uiBlockSetCol(iBlock, TH_AUTO);

	char* lStr = "Value %t|False %x0| True%x1";
	uiDefButS ( iBlock, MENU, B_MATPRV_DRAW, lStr
						, aM->iX, aM->iY-20, aM->iSizeX, 20
						, &aM->i, 0.0, 0.0, 0, 0
						, "Value of the constant boolean. Click to change.");

}


/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////


int BlenderDrawVisitor::PopupMenu( const Options& aOptions )
{
	std::stringstream lMenu;
	
	int i=1;
	for ( Options::const_iterator it = aOptions.begin()
		; it != aOptions.end()
		; it++, i++ )
		{
			if (i>1) lMenu<<"|";
			lMenu << (*it) << "%x"<<i;
		}
	
	int val = pupmenu( const_cast<char*>( lMenu.str().c_str()) );
	
	if ( (val>0) && (val<(int)aOptions.size()+1) ) return val-1;
	else return -1;
}


