

#ifndef BLENDERDRAW_H_
#define BLENDERDRAW_H_

#include "Module.h"
#include "ModuleDraw.h"

#include "modules/generic/EndOfData.h"
#include "modules/generic/Export.h"
#include "modules/generic/If.h"
#include "modules/generic/Import.h"
#include "modules/generic/MarkGroup.h"
#include "modules/generic/Replicate.h"
#include "modules/generic/SequenceSelect.h"

#include "modules/blender/SelectBlenderScene.h"
#include "modules/blender/BlenderObjectFilter.h"

#include "modules/logic/ConstantReal.h"
#include "modules/logic/VariableReal.h"
#include "modules/logic/AddReal.h"
#include "modules/logic/CompareReal.h"

#include "modules/logic/ConstantVector.h"
#include "modules/logic/VariableVector.h"

#include "modules/logic/ConstantBoolean.h"
#include "modules/logic/VariableBoolean.h"
#include "modules/logic/OperateBoolean.h"

#include "modules/logic/ConstantInteger.h"
#include "modules/logic/VariableInteger.h"

#include "modules/logic/KeyEvent.h"
#include "modules/logic/EnablePipeline.h"
#include "modules/logic/PhysicsModifier.h"
#include "modules/logic/SelectObject.h"


extern "C" {

#include "BIF_gl.h"
#include "BIF_resources.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "BIF_butspace.h"
#include "BIF_screen.h"
#include "BIF_keyval.h"
#include "BIF_mywindow.h"
#include "PIL_time.h"

#include "DNA_object_types.h"
#include "DNA_screen_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"

#include "butspace.h"
#include "interface.h"
#include "mydevice.h"

}


#include <string>
#include <vector>
#include <sstream>


namespace grn 
{

	class BlenderDrawVisitor 
		: public ModuleDrawVisitor
		
		, public Visitor<EndOfData>
		, public Visitor<Export>
		, public Visitor<If>
		, public Visitor<Import>
		, public Visitor<MarkGroup>
		, public Visitor<Replicate>
		, public Visitor<SequenceSelect>
		
		, public Visitor<SelectBlenderScene>
		, public Visitor<BlenderObjectFilter>
		
		, public Visitor<ConstantReal>
		, public Visitor<CompareReal>
		, public Visitor<VariableReal>
		, public Visitor<AddReal>

		, public Visitor<ConstantBoolean>
		, public Visitor<VariableBoolean>
		, public Visitor<OperateBoolean>

		, public Visitor<ConstantInteger>
		, public Visitor<VariableInteger>

		, public Visitor<ConstantVector>
		, public Visitor<VariableVector>

		, public Visitor<KeyEvent>
		, public Visitor<EnablePipeline>
		, public Visitor<PhysicsModifier>
		, public Visitor<SelectObject>
	{
	public:
		
		uiBlock *iBlock;
		
	public:
	
		void Visit( Module* aM );
	
		/** @{
		 * Generic modules 
		 */
		void Visit( EndOfData* aM );
		void Visit( Export* aM );
		void Visit( If* aM );
		void Visit( Import* aM );
		void Visit( MarkGroup* aM );
		void Visit( Replicate* aM );
		void Visit( SequenceSelect* aM );
		/** @} */
		
		
		/** @{
		 * Blender modules 
		 */
		void Visit( SelectBlenderScene* aM );
		void Visit( BlenderObjectFilter* aM );
		/** @} */

		
		/** @{
		 * Logic modules 
		 */
		void Visit( ConstantBoolean* aM );
		void Visit( VariableBoolean* aM );
		void Visit( OperateBoolean* aM );

		void Visit( ConstantInteger* aM );
		void Visit( VariableInteger* aM );

		void Visit( ConstantVector* aM );
		void Visit( VariableVector* aM );
		
		void Visit( ConstantReal* aM );
		void Visit( VariableReal* aM );
		void Visit( AddReal* aM );
		void Visit( CompareReal* aM );

		void Visit( KeyEvent* aM );
		void Visit( EnablePipeline* aM );
		void Visit( PhysicsModifier* aM );
		void Visit( SelectObject* aM );
		/** @} */
		
	protected:

		/** Helper methods for common UI functionality 
		 */
		typedef std::vector< std::string > Options;
		static int PopupMenu( const Options& aOptions );
			
	};

}


#endif /*BLENDERDRAW_H_*/

