
#include "BlenderRootPipeline.h"

#include "datatypes/BlenderTypes.h"

#include "modules/blender/BlenderSource.h"
#include "modules/blender/BlenderSceneParts.h"
#include "modules/blender/BlenderObjectFilter.h"
#include "modules/blender/BlenderObjectParts.h"
#include "modules/blender/SelectBlenderScene.h"

#include "modules/MeshConvert.h"
#include "modules/MaterialConvert.h"
#include "modules/CameraConvert.h"
#include "modules/SpotlightConvert.h"

#include "GironaLog.h"

using namespace grn;

BlenderRootPipeline::BlenderRootPipeline()
{
	// Register data types	
	GRN_RegisterDataType( BlenderScene );
	GRN_RegisterDataType( BlenderObject );
	GRN_RegisterDataType( BlenderMesh );
	GRN_RegisterDataType( BlenderMaterial );
	
	
	// Register modules	
	iModuleFactories.push_back( new BlenderSource::Factory() );
	iModuleFactories.push_back( new BlenderSceneParts::Factory() );
	iModuleFactories.push_back( new SelectBlenderScene::Factory() );
	iModuleFactories.push_back( new SpotlightConvert::Factory() );
	iModuleFactories.push_back( new MeshConvert::Factory() );
	iModuleFactories.push_back( new MaterialConvert::Factory() );
	iModuleFactories.push_back( new CameraConvert::Factory() );
	iModuleFactories.push_back( new BlenderObjectFilter::Factory() );
	iModuleFactories.push_back( new BlenderObjectParts::Factory() );

	
}

