

#include "girona.h"

extern "C" {

#include "BIF_gl.h"
#include "BIF_resources.h"
#include "BIF_space.h"
#include "BIF_interface.h"
#include "BIF_butspace.h"
#include "BIF_screen.h"
#include "BIF_keyval.h"
#include "BIF_mywindow.h"
#include "PIL_time.h"

#include "DNA_object_types.h"
#include "DNA_screen_types.h"
#include "DNA_space_types.h"
#include "DNA_scene_types.h"

#include "butspace.h"
#include "interface.h"
#include "mydevice.h"

}


#include "Module.h"
#include "Pipeline.h"
#include "BlenderRootPipeline.h"
#include "BlenderDraw.h"
#include "BlenderEngine.h"
#include "girona/engine/GL2/RendererGL2.h"


#include <sstream>

using namespace grn;

static RootPipeline* gRootPipeline = 0;

static short int gCurrentPipeline = 0;
static short int gExecutionStage = 0;
static short int gExecutionOrder = 0;
static short int gStartEnabled = 0;
static char gPipelineName[36];

RootPipeline* get_root_pipeline() { return gRootPipeline; }


static void add_module(void *aPipeline, void *)
{
	Pipeline* lPipeline = static_cast<Pipeline*>( aPipeline );
	
	if (!lPipeline) return;

	Module *lNewModule = 0;
	
	if (aPipeline!=&gRootPipeline)
	{
		std::stringstream lMenu;
		
		int i=1;
		for ( Pipeline::ModuleFactoryList::iterator it = gRootPipeline->iModuleFactories.begin()
			; it != gRootPipeline->iModuleFactories.end()
			; it++, i++ )
			{
				if (i>1) lMenu<<"|";
				lMenu << (*it)->getName() << "%x"<<i;
			}
		
		int val = pupmenu( const_cast<char*>( lMenu.str().c_str()) );
		
		if ( (val>0) && (val<(int)gRootPipeline->iModuleFactories.size()+1) )
		{
			lNewModule = gRootPipeline->iModuleFactories[val-1]->New( lPipeline );
			lPipeline->iModules.push_back( lNewModule );
		}
	}
	else
	{
		lNewModule = lPipeline->FindFactory(Pipeline::sName)->New( lPipeline );
		lPipeline->iModules.push_back( lNewModule );
	}
	
	// Select the newly created module
	lPipeline->iSelected = lNewModule;
	
}


static void del_module(void* aPipeline, void* )
{
	Pipeline* lPipeline = static_cast<Pipeline*>( aPipeline );
	lPipeline->RemoveSelected();
}


static void change_pipeline_name(void *aPipeline, void *)
{
	Pipeline* lPipeline = static_cast<Pipeline*>( aPipeline );
	lPipeline->SetName( gPipelineName );

	allqueue(REDRAWBUTSLOGIC, curarea->win);	
}


static void save_pipeline(void *aPipeline, void *)
{
	Pipeline* lPipeline = static_cast<Pipeline*>( aPipeline );
	lPipeline->Save( lPipeline->GetName()+".pipeline" );
}


static void load_pipeline(void *aPipeline, void *)
{
	Pipeline* lPipeline = static_cast<Pipeline*>( aPipeline );
	lPipeline->Load( lPipeline->GetName()+".pipeline" );
}

static void build_data(void *, void *)
{
	// create an engine
	Renderer* lRenderer = new grn::RendererGL2();
	BlenderEngine* lEngine = new BlenderEngine(gRootPipeline, lRenderer);

	lEngine->Build( gRootPipeline->GetName() + ".data" );	
	gRootPipeline->SaveRuntime( gRootPipeline->GetName() + ".girona" );	
	
	// deallocate
	delete lEngine;
	delete lRenderer;
	
}

Pipeline* get_pipeline()
{
	Pipeline* lPipeline = 0;
	
	if  ((gCurrentPipeline-1>=0) && (gCurrentPipeline-1<(int)gRootPipeline->iModules.size()))
		lPipeline = dynamic_cast<Pipeline*>(gRootPipeline->iModules[ gCurrentPipeline-1 ]);
	else lPipeline = gRootPipeline;
	
	return lPipeline;
}


static void select_execution_stage(void *aP, void *)
{
	RootPipeline::ExecutionStage lStage = RootPipeline::eNone;
	switch (gExecutionStage)
	{
		case 1 : lStage = RootPipeline::eBuild; break;
		case 2 : lStage = RootPipeline::eLogic; break;
		case 3 : lStage = RootPipeline::eRender; break;
		default: lStage = RootPipeline::eNone;
	}
	gRootPipeline->SetExecutionStage( static_cast<Pipeline*>(aP), lStage, gExecutionOrder );
}


static void set_start_enabled(void *aP, void *)
{
	Pipeline* lP = static_cast<Pipeline*>(aP);
	lP->SetStartEnabled(gStartEnabled);
}


//
//	Draw the pipeline interface
//

void girona_draw_interface()
{
	// Dirty create
	if (!gRootPipeline) gRootPipeline = new BlenderRootPipeline();
	
	Pipeline* lPipeline = get_pipeline();
	
	// Root pipeline extra interface
	if (lPipeline->GetParent()==gRootPipeline)
	{
		char name[32];	
		sprintf(name, "root_pipeline %d", curarea->win);
		uiBlock *block= uiNewBlock(&curarea->uiblocks, name, UI_EMBOSS, UI_HELV, curarea->win);
		RootPipeline::ExecutionStage lStage = gRootPipeline->GetExecutionStage(lPipeline);
		switch (lStage)
		{
			case RootPipeline::eBuild: gExecutionStage = 1; break;
			case RootPipeline::eLogic: gExecutionStage = 2; break;
			case RootPipeline::eRender: gExecutionStage = 3; break;
			default: gExecutionStage = 0;
		}
		std::string lMenu = "Execution stage%t|Not executed %x0|Build stage%x1|Logic stage%x2|Render stage%x3";
		
		uiBut* iBase = uiDefButS ( block, MENU, 1, const_cast<char*>( lMenu.c_str() )
								, 0, 210, 110, 24
								, &gExecutionStage, 0.0, 0.0, 0, 0
								, "Selects the execution stage for the currently editing pipeline.");
		uiButSetFunc( iBase, select_execution_stage, lPipeline, NULL);

		gExecutionOrder = gRootPipeline->GetExecutionOrder(lPipeline);
		iBase = uiDefButS( block, NUM|INT, 1, ""
								, 111, 210, 60, 24
								, &gExecutionOrder, -100.0, 100.0, 1, 0
								, "Selects the execution order inside the stage for the currently editing pipeline.");
		uiButSetFunc( iBase, select_execution_stage, lPipeline, NULL);

		gStartEnabled = lPipeline->GetStartEnabled();
		iBase = uiDefButS ( block, TOG, 1, "Start Enabled"
								, 0, 185, 170, 24
								, &gStartEnabled, 0.0, 0.0, 0, 0
								, "Selects whteher this pipeline starts enabled or not.");
		uiButSetFunc( iBase, set_start_enabled, lPipeline, NULL);


		strncpy(gPipelineName,lPipeline->GetName().c_str(),31);
		iBase = uiDefBut( block, TEX, 1, "Name:"
						, 0, 235, 170, 24
						, &gPipelineName, 0, 31, 0, 0
						, "Name of the current pipeline.");
		uiButSetFunc( iBase, change_pipeline_name, lPipeline, NULL);

		uiBlockSetCol(block, TH_BUT_SETTING);
		iBase = uiDefBut( block, BUT, 1, "Add Module...",		
						  0,160, 170,24,
						  NULL, 0.0, 100.0, 100, 0,
						  "");
		uiButSetFunc(iBase, add_module, lPipeline, NULL);
	
		iBase = uiDefBut( block, BUT, 1, "Delete Selected",
						  0,135, 170,24,
						  NULL, 0.0, 100.0, 100, 0,
						  "");
		uiButSetFunc(iBase, del_module, lPipeline, NULL);
	
		uiComposeLinks(block);
		uiDrawBlock(block);	
	}
	else if (lPipeline==gRootPipeline)
	{
		char name[32];	
		sprintf(name, "root_pipeline %d", curarea->win);
		uiBlock *block= uiNewBlock(&curarea->uiblocks, name, UI_EMBOSS, UI_HELV, curarea->win);
	
		uiBlockSetCol(block, TH_BUT_SETTING);
		uiBut* iBase = uiDefBut( block, BUT, 1, "Add Pipeline...",		
						  0,160, 170,24,
						  NULL, 0.0, 100.0, 100, 0,
						  "");
		uiButSetFunc(iBase, add_module, lPipeline, NULL);
	
		iBase = uiDefBut( block, BUT, 1, "Delete Selected",
						  0,135, 170,24,
						  NULL, 0.0, 100.0, 100, 0,
						  "");
		uiButSetFunc(iBase, del_module, lPipeline, NULL);
	
		iBase = uiDefBut( block, BUT, 1, "Build data",
						  0,195, 170,24,
						  NULL, 0.0, 100.0, 100, 0,
						  "");
		uiButSetFunc(iBase, build_data, lPipeline, NULL);
	
		uiComposeLinks(block);
		uiDrawBlock(block);	
	}
	
	// Common interface
	char name[32];	
	sprintf(name, "common_pipeline %d", curarea->win);
	uiBlock *block= uiNewBlock(&curarea->uiblocks, name, UI_EMBOSS, UI_HELV, curarea->win);

	uiBut* iBase = uiDefBut( block, BUT, 1, "Save Pipeline",
					  0,10, 170,24,
					  NULL, 0.0, 100.0, 100, 0,
					  "");
	uiButSetFunc(iBase, save_pipeline, lPipeline, NULL);

	iBase = uiDefBut( block, BUT, 1, "Load Pipeline",		
					  0,35, 170,24,
					  NULL, 0.0, 100.0, 100, 0,
					  "");
	uiButSetFunc(iBase, load_pipeline, lPipeline, NULL);

	std::stringstream lMenu;
	lMenu<<"Pipelines%t|Root %x0";
	for ( unsigned int i=0; i<gRootPipeline->iModules.size(); i++ ) 
	{
		Pipeline* lP = dynamic_cast<Pipeline*>( gRootPipeline->iModules[i] );
		if (lP) lMenu<<"| "<<lP->GetName()<<" %x"<<i+1;
	}
	
	iBase = uiDefButS ( block, MENU, B_MATPRV_DRAW, const_cast<char*>( lMenu.str().c_str() )
				, 0, 60, 170, 24
				, &gCurrentPipeline, 0.0, 0.0, 0, 0
				, "Selects the pipeline to open in the editor.");

	uiComposeLinks(block);
	uiDrawBlock(block);	
	
	// Pipeline layout
	sprintf(name, "pipeline_itslef %d", curarea->win);
	block= uiNewBlock(&curarea->uiblocks, name, UI_EMBOSS, UI_HELV, curarea->win);
	
	BlenderDrawVisitor lDrawer;
	lDrawer.iBlock = block;

	lPipeline->Draw(lDrawer);
	
	uiComposeLinks(block);
	uiDrawBlock(block);	
}


//
// Moving of modules around
//

static bool sMoving;
static short sMousePos[2];

void girona_move_interface( short val )
{
	if (!get_pipeline()) return;

	if ((!sMoving) && (val) )
	{
		sMoving = true;
		
		// get previous mouse position
		getmouseco_areawin(sMousePos);
	}
	
	if (!val) sMoving = false;

}



int girona_mouse_interface()
{
	Pipeline* lPipeline = get_pipeline();
	if ( (!lPipeline) || (!sMoving) ) return 0;

	if (lPipeline->iSelected)
	{
		// get new mouse position
		short lNewPos[2];
		getmouseco_areawin(lNewPos);
		
		lPipeline->iSelected->iX += lNewPos[0]-sMousePos[0];
		lPipeline->iSelected->iY += lNewPos[1]-sMousePos[1];
		
		sMousePos[0] = lNewPos[0];
		sMousePos[1] = lNewPos[1];
		
		return 1;
	}
	
	return 0;
}

