
#include "MaterialConvert.h"

#include "Pipeline.h"

using namespace grn;


IMPLEMENT_MODULE(MaterialConvert);


MaterialConvert::MaterialConvert( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,50);
	iX = 250;
	rInput = new Module::tInputPad<grn::BlenderMaterial>( "BlenderMaterial", this );
	rOutput = new Module::tOutputPad<grn::Material>( "Material", this );
}


void MaterialConvert::Run()
{
	while (rInput->hasData())
	{
		Pointer<BlenderMaterial> lInMaterial = rInput->getData( true );
		
		Pointer<Material> lMat = 0;

		if (lInMaterial->rMaterial)
		{
			::Material* rM = lInMaterial->rMaterial;
			// Convert material, phong case
			Pointer<PhongMaterial> lOutMat = new PhongMaterial( rM->mode & MA_SHADOW );
			lOutMat->iAmbient.iColor = SimdVector4( rM->ambr, rM->ambg, rM->ambb, 1.0f );
			lOutMat->iDiffuse.iColor = SimdVector4( rM->r, rM->g, rM->b, 1.0f );
			lOutMat->iSpecular.iColor = SimdVector4( rM->specr, rM->specg, rM->specb, 1.0f );
			lOutMat->iExponent = rM->spec;
			lMat = lOutMat;
		}
		else
		{
			// Send a default nice multicolored phong material
			Pointer<PhongMaterial> lOutMat = new PhongMaterial( true );
			lOutMat->iAmbient.iColor = SimdVector4(0.1f,0.5f,0.1f,1.0f);
			lOutMat->iDiffuse.iColor = SimdVector4(1.0f,0.0f,0.0f,1.0f);
			lOutMat->iSpecular.iColor = SimdVector4(1.0f,1.0f,1.0f,1.0f);
			lOutMat->iExponent = 2.0f;
			
			lMat = lOutMat;
		}

		rPipeline->Emit(rOutput,lMat);
	}
}
