
#include "SpotlightConvert.h"

#include "Pipeline.h"

#include "DNA_lamp_types.h"

using namespace grn;


IMPLEMENT_MODULE(SpotlightConvert);


SpotlightConvert::SpotlightConvert( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,50);
	iX = 250;
	rInput = new Module::tInputPad<grn::BlenderObject>( "BlenderObject", this );
	rOutput = new Module::tOutputPad<grn::SpotLight>( "SpotLight", this );
}


void SpotlightConvert::Run()
{
	while (rInput->hasData())
	{
		Pointer<BlenderObject> lInObject = rInput->getData( true );
		Pointer<Transform> lTrans = lInObject->ExtractTransform();

		// Emit subdata if suitable types
		::Object* ob = lInObject->rObject;
		if (ob->type == OB_LAMP)
		{
			::Lamp* lamp = static_cast<Lamp*>(ob->data);
			
			if (lamp->type==LA_SPOT)
			{
				Pointer<SpotLight> lSpot = new SpotLight();
				lSpot->iColor = SimdVector3( lamp->r,lamp->g,lamp->b );
				lSpot->iTransform = lInObject->ExtractTransform();
				lSpot->iAngle = atan2f(lamp->spotsize/2.0f, lamp->dist) * 2.0f;
				
				rPipeline->Emit(rOutput,lSpot);
			}
			
		}
		
	}
}

