
#include "BlenderObjectFilter.h"

#include "Pipeline.h"

/* This list includes only data type definitions */
#include "DNA_object_types.h"
#include "DNA_material_types.h"
#include "DNA_image_types.h"
#include "DNA_lamp_types.h"
#include "DNA_group_types.h"
#include "DNA_scene_types.h"
#include "DNA_camera_types.h"
#include "DNA_property_types.h"
#include "DNA_text_types.h"
#include "DNA_sensor_types.h"
#include "DNA_controller_types.h"
#include "DNA_actuator_types.h"
#include "DNA_mesh_types.h"
#include "DNA_meshdata_types.h"
#include "DNA_view3d_types.h"
#include "DNA_world_types.h"
#include "DNA_sound_types.h"
#include "DNA_key_types.h"
#include "DNA_armature_types.h"


using namespace grn;

IMPLEMENT_MODULE(BlenderObjectFilter);


BlenderObjectFilter::BlenderObjectFilter( Pipeline* aParent )
: Module(aParent)
{
	setSize(160,30);
	iX = 300;
	iY = 150;
	rInput = new Module::tInputPad<grn::BlenderObject>( "InputObjects", this );
	rOutput = new Module::tOutputPad<grn::BlenderObject>( "OutputObjects", this );
}


void BlenderObjectFilter::Run()
{
	while (rInput->hasData())
	{
		Pointer<BlenderObject> lInObj = rInput->getData(true);
		
		// Check types according to the enum in the string od the Draw function.
		if (   ( (iType==0) && (lInObj->rObject->type == OB_MESH) )
			|| ( (iType==1) && (lInObj->rObject->type == OB_ARMATURE) )
			|| ( (iType==2) && (lInObj->rObject->type == OB_CAMERA) )
			|| ( (iType==3) && (lInObj->rObject->type == OB_LAMP) && ( static_cast<Lamp*>(lInObj->rObject->data)->type == LA_SPOT) )
			)
		rPipeline->Emit(rOutput,lInObj);
	}
}


void BlenderObjectFilter::Serialize( std::ostream& s ) const
{
	// Serialize parent first
	Module::Serialize(s);
	
	s << iType << std::endl;
}


void BlenderObjectFilter::Deserialize( std::istream& s )
{
	// Deserialize parent first
	Module::Deserialize(s);

	s>>iType;
}

