
#include "BlenderObjectParts.h"

#include "Pipeline.h"

#include "datatypes/Object.h"

#include "MT_Point3.h"
#include "MT_Vector3.h"
#include "MT_Matrix3x3.h"
#include "MT_Transform.h"

#include "MT_MinMax.h"
#include "MEM_guardedalloc.h"
#include "BKE_utildefines.h"
#include "BKE_material.h"

#include "DNA_property_types.h"
#include "DNA_key_types.h"

using namespace grn;


IMPLEMENT_MODULE(BlenderObjectParts);


BlenderObjectParts::BlenderObjectParts( Pipeline* aParent )
: Module(aParent)
{
	setSize(140,160);
	iX = 300;
	iY = 150;
	rInput = new Module::tInputPad<grn::BlenderObject>( "InputObjects", this );
	roMesh = new Module::tOutputPad<grn::BlenderMesh>( "Mesh", this );
	roMaterial = new Module::tOutputPad<grn::BlenderMaterial>( "Material", this );
	roTransform = new Module::tOutputPad<grn::Transform>( "Transform", this );
	roName = new Module::tOutputPad<grn::String>( "Name", this );
	roParentName = new Module::tOutputPad<grn::String>( "ParentName", this );
	roProperties = new Module::tOutputPad<grn::StringList>( "Properties", this );
	roPhysics = new Module::tOutputPad<grn::PhysicsProperties>( "PhysicsProperties", this );
}



////////////////////////////////////////////
/////////@TODO: Cleanup this imported code
////////////////////////////////////////////


static float my_boundbox_mesh(::Mesh *me, float *loc, float *size)
{
	MVert *mvert;
	BoundBox *bb;
	MT_Point3 min, max;
	float mloc[3], msize[3];
	int a;
	
	if(me->bb==0) me->bb= (struct BoundBox *)MEM_callocN(sizeof(BoundBox), "boundbox");
	bb= me->bb;
	
	INIT_MINMAX(min, max);

	if (!loc) loc= mloc;
	if (!size) size= msize;
	
	mvert= me->mvert;
	for(a=0; a<me->totvert; a++, mvert++) {
		DO_MINMAX(mvert->co, min, max);
	}
		
	if(me->totvert) {
		loc[0]= (min[0]+max[0])/2.0;
		loc[1]= (min[1]+max[1])/2.0;
		loc[2]= (min[2]+max[2])/2.0;
		
		size[0]= (max[0]-min[0])/2.0;
		size[1]= (max[1]-min[1])/2.0;
		size[2]= (max[2]-min[2])/2.0;
	}
	else {
		loc[0]= loc[1]= loc[2]= 0.0;
		size[0]= size[1]= size[2]= 0.0;
	}
		
	bb->vec[0][0]=bb->vec[1][0]=bb->vec[2][0]=bb->vec[3][0]= loc[0]-size[0];
	bb->vec[4][0]=bb->vec[5][0]=bb->vec[6][0]=bb->vec[7][0]= loc[0]+size[0];
		
	bb->vec[0][1]=bb->vec[1][1]=bb->vec[4][1]=bb->vec[5][1]= loc[1]-size[1];
	bb->vec[2][1]=bb->vec[3][1]=bb->vec[6][1]=bb->vec[7][1]= loc[1]+size[1];

	bb->vec[0][2]=bb->vec[3][2]=bb->vec[4][2]=bb->vec[7][2]= loc[2]-size[2];
	bb->vec[1][2]=bb->vec[2][2]=bb->vec[5][2]=bb->vec[6][2]= loc[2]+size[2];

	float radius = 0;
	for (a=0, mvert = me->mvert; a < me->totvert; a++, mvert++)
	{
		float vert_radius = MT_Vector3(mvert->co).length2();
		if (vert_radius > radius)
			radius = vert_radius;
	} 
	return sqrt(radius);
}

static void my_tex_space_mesh(::Mesh *me)
{
	KeyBlock *kb;
	float *fp, loc[3], size[3], min[3], max[3];
	int a;

	my_boundbox_mesh(me, loc, size);
	
	if(me->texflag & AUTOSPACE) {
		if(me->key) {
			kb= me->key->refkey;
			if (kb) {
	
				INIT_MINMAX(min, max);
		
				fp= (float *)kb->data;
				for(a=0; a<kb->totelem; a++, fp+=3) {	
					DO_MINMAX(fp, min, max);
				}
				if(kb->totelem) {
					loc[0]= (min[0]+max[0])/2.0; loc[1]= (min[1]+max[1])/2.0; loc[2]= (min[2]+max[2])/2.0;
					size[0]= (max[0]-min[0])/2.0; size[1]= (max[1]-min[1])/2.0; size[2]= (max[2]-min[2])/2.0;
	} 
	else {
					loc[0]= loc[1]= loc[2]= 0.0;
					size[0]= size[1]= size[2]= 0.0;
				}
				
			}
				}
	
		VECCOPY(me->loc, loc);
		VECCOPY(me->size, size);
		me->rot[0]= me->rot[1]= me->rot[2]= 0.0;
	
		if(me->size[0]==0.0) me->size[0]= 1.0;
		else if(me->size[0]>0.0 && me->size[0]<0.00001) me->size[0]= 0.00001;
		else if(me->size[0]<0.0 && me->size[0]> -0.00001) me->size[0]= -0.00001;
	
		if(me->size[1]==0.0) me->size[1]= 1.0;
		else if(me->size[1]>0.0 && me->size[1]<0.00001) me->size[1]= 0.00001;
		else if(me->size[1]<0.0 && me->size[1]> -0.00001) me->size[1]= -0.00001;
						
		if(me->size[2]==0.0) me->size[2]= 1.0;
		else if(me->size[2]>0.0 && me->size[2]<0.00001) me->size[2]= 0.00001;
		else if(me->size[2]<0.0 && me->size[2]> -0.00001) me->size[2]= -0.00001;
	}
	
}

static void my_get_local_bounds(::Object *ob, float *centre, float *size)
{
	BoundBox *bb= NULL;
	// uses boundbox, function used by Ketsji
	switch (ob->type)
	{
		case OB_MESH:
			bb= ( (::Mesh *)ob->data )->bb;
			if(bb==0) 
			{
				my_tex_space_mesh((struct ::Mesh *)ob->data);
				bb= ( (::Mesh *)ob->data )->bb;
			}
			break;
		case OB_CURVE:
		case OB_SURF:
		case OB_FONT:
			centre[0]= centre[1]= centre[2]= 0.0;
			size[0]  = size[1]=size[2]=0.0;
			break;
		case OB_MBALL:
			bb= ob->bb;
			break;
	}
	
	if(bb==NULL) 
	{
		centre[0]= centre[1]= centre[2]= 0.0;
		size[0] = size[1]=size[2]=1.0;
	}
	else 
	{
		size[0]= 0.5*fabs(bb->vec[0][0] - bb->vec[4][0]);
		size[1]= 0.5*fabs(bb->vec[0][1] - bb->vec[2][1]);
		size[2]= 0.5*fabs(bb->vec[0][2] - bb->vec[1][2]);
					
		centre[0]= 0.5*(bb->vec[0][0] + bb->vec[4][0]);
		centre[1]= 0.5*(bb->vec[0][1] + bb->vec[2][1]);
		centre[2]= 0.5*(bb->vec[0][2] + bb->vec[1][2]);
	}
}


////////////////////////////////////////////


void BlenderObjectParts::Run()
{
	while (rInput->hasData())
	{
		Pointer<BlenderObject> lInObj = rInput->getData(true);
		
		// mesh
		if (lInObj->rObject->type==OB_MESH)
		{
			Pointer<BlenderMesh> lMesh = new BlenderMesh();
			lMesh->rMesh = static_cast< ::Mesh* >(lInObj->rObject->data);
			rPipeline->Emit(roMesh,lMesh);
		}
		
		// transform
		{
			Pointer<Transform> lTrans = lInObj->ExtractTransform();
			rPipeline->Emit(roTransform,lTrans);
		}
		
		// name
		{
			Pointer<String> lName = new String();
			lName->iValue = lInObj->rObject->id.name;
			rPipeline->Emit(roName,lName);
		}
		
		// parent name
		{
			Pointer<String> lName = new String();
			if (lInObj->rObject->parent)
				lName->iValue = lInObj->rObject->parent->id.name;
			else lName->iValue = kNullObjectName;
			rPipeline->Emit(roParentName,lName);
		}
		
		// material
		{
			Pointer<BlenderMaterial> lMaterial = new BlenderMaterial();
			
			if (lInObj->rObject->mat)
				lMaterial->rMaterial = lInObj->rObject->mat[0];
			else lMaterial->rMaterial = 0;
			
			rPipeline->Emit(roMaterial,lMaterial);
		}
		
		// user properties
		{
			Pointer<StringList> lProps = new StringList();

			for (bProperty* p = (bProperty*)lInObj->rObject->prop.first; p; p= (bProperty*) p->next)
				lProps->iValue.push_back(p->name);

			rPipeline->Emit(roProperties,lProps);
		}
		
		// physics properties
		{
			Pointer<PhysicsProperties> lProps = new PhysicsProperties();

			lProps->iActor = ((lInObj->rObject->gameflag & OB_ACTOR)!=0);
			lProps->iDynamic =  ((lInObj->rObject->gameflag & OB_DYNAMIC) != 0);
			lProps->iRigidBody = ((lInObj->rObject->gameflag & OB_RIGID_BODY) != 0);
			lProps->iGhost = ((lInObj->rObject->gameflag & OB_GHOST) != 0);
			lProps->iDeformable = (((lInObj->rObject->gameflag2 & 2)) != 0);

			::Material* blendermat = give_current_material(lInObj->rObject, 0);
			if (blendermat)
			{
				lProps->iFriction = blendermat->friction;
				lProps->iRestitution = blendermat->reflect;
			}
			else
			{
				lProps->iFriction = 2.0f;
				lProps->iRestitution = 1.0f;
			}
			 
			lProps->iMass = lInObj->rObject->mass;
			lProps->iInertia = lInObj->rObject->formfactor;
			lProps->iRadius = lInObj->rObject->inertia;		// this is not a bug, or it is wrong in bullet-ketsji also.
			lProps->iLinearDrag = lInObj->rObject->damping;
			lProps->iAngularDrag = lInObj->rObject->rdamping;
			lProps->iHasAnisotropicFriction = ((lInObj->rObject->gameflag & OB_ANISOTROPIC_FRICTION) != 0);
			lProps->iAnisotropicFrictionScaling = SimdVector3( lInObj->rObject->anisotropicFriction[0], lInObj->rObject->anisotropicFriction[1], lInObj->rObject->anisotropicFriction[2] );
			lProps->iDoFH = ((lInObj->rObject->gameflag & OB_DO_FH) != 0);
			lProps->iDoRotFH = ((lInObj->rObject->gameflag & OB_ROT_FH) != 0);
		
			lProps->iLocalBoundsCenter;
			lProps->iLocalBoundsExtends;
		
			switch (lInObj->rObject->boundtype)
			{
				case OB_BOUND_BOX: lProps->iBoundType = PhysicsProperties::eBoundBox; break;
				case OB_BOUND_POLYT: lProps->iBoundType = PhysicsProperties::eBoundPolyt; break;
				case OB_BOUND_POLYH: lProps->iBoundType = PhysicsProperties::eBoundPolyh; break;
				case OB_BOUND_SPHERE: lProps->iBoundType = PhysicsProperties::eBoundSphere; break;
				case OB_BOUND_CYLINDER: lProps->iBoundType = PhysicsProperties::eBoundCylinder; break;
				case OB_BOUND_CONE: lProps->iBoundType = PhysicsProperties::eBoundCone; break;
			}
			
			lProps->iConcave = ((lInObj->rObject->boundtype & 4) != 0);


			float bbm_center[3], bbm_extends[3];
			my_get_local_bounds(lInObj->rObject,bbm_center,bbm_extends);
			lProps->iLocalBoundsCenter[0] = bbm_center[0];
			lProps->iLocalBoundsCenter[1] = bbm_center[1];
			lProps->iLocalBoundsCenter[2] = bbm_center[2];
			lProps->iLocalBoundsExtends[0] = bbm_extends[0];
			lProps->iLocalBoundsExtends[1] = bbm_extends[1];
			lProps->iLocalBoundsExtends[2] = bbm_extends[2];

			rPipeline->Emit(roPhysics,lProps);
		}
		
	}
}


