
/*
#include <GL/glut.h>
#include <cstdio>
#include <cstdlib>

#ifdef _WIN32
#	include <windows.h>
#else
#	include <sys/time.h>
#endif

#include "PlayerEngine.h"

#include "girona/RootPipeline.h"
#include "girona/Serialise.h"

#include <pspmoduleinfo.h>


PSP_MODULE_INFO("GironaPlayer", 0, 1, 1);



static int g_Width = 512;
static int g_Height = 512;


#ifdef _WIN32
static DWORD last_idle_time;
#else
static struct timeval last_idle_time;
#endif



static PlayerEngine* gEngine(0);
static grn::RootPipeline* gRoot(0);
static bool gExit(false);


void display(void)
{
	float lDeltaTime;
	
#ifdef _WIN32
	DWORD time_now;
	time_now = GetTickCount();
	lDeltaTime = (float) (time_now - last_idle_time) / 1000.0;
#else
	// Figure out time elapsed since last call to idle function
	struct timeval time_now;
	gettimeofday(&time_now, NULL);
	lDeltaTime = (float)(time_now.tv_sec  - last_idle_time.tv_sec) + 1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
#endif

	
//	double lCurrentTime = PIL_check_seconds_timer();
	double lCurrentTime = 0.0;
	
	// Tell the engine the delta time
	gEngine->SetDeltaTime(lDeltaTime);
	gEngine->SetCurrentTime(lCurrentTime);

	// Run a frame
	gEngine->RunFrame();

	// Save time_now for next time
	last_idle_time = time_now;
}


void Keyboard(unsigned char key, int , int )
{
#ifndef GIRONA_PSP	
	if (key==27) exit(0);
#else
	if (key=='x') exit(0);
#endif
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], true);
}

void KeyboardUp(unsigned char key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], false);
}


void SpecialKeyboard(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], true);
}

void SpecialKeyboardUp(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], false);
}


void loop()
{
	// Force redraw
	glutPostRedisplay();
	
}

 
int main(int argc, char** argv)
{
	// GLUT Window Initialization:
	glutInit (&argc, argv);
	glutInitWindowSize (g_Width, g_Height);
	glutInitDisplayMode ( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow ("Blender-Girona Player");

	// Initialize the game logic
	gRoot = new grn::RootPipeline();
	gRoot->Load("unnamed.girona");
	
	// Create engine class
	gEngine = new PlayerEngine( gRoot );
	
	// Load game data
	gEngine->LoadData( "unnamed.data" );

	// Set up the game render pipeline
	gEngine->SetScreenArea( 0, 0, g_Width, g_Height );
	gEngine->Start();
	
	gExit = false;

	// Register callbacks:
	glutDisplayFunc (display);
	glutReshapeFunc (reshape);
	glutKeyboardFunc (Keyboard);
	glutKeyboardUpFunc (KeyboardUp);
	glutSpecialFunc (SpecialKeyboard);
	glutSpecialUpFunc (SpecialKeyboardUp);
	//glutMouseFunc (MouseButton);
	//glutMotionFunc (MouseMotion);
	glutIdleFunc (loop);


	// Get the initial time
	gettimeofday (&last_idle_time, NULL);

	// Turn the flow of control over to GLUT
	glutMainLoop ();
		
	gRoot->ClearData();
		
	delete gEngine;

	delete gRoot;
		
	return 1;
	
}

*/


#include <stdlib.h>
#include <GL/glut.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>

#ifdef GIRONA_PSP
	#include <pspkernel.h>
	#include <pspdisplay.h>
	#include <pspdebug.h>
	PSP_MODULE_INFO("GironaPlayer", 0, 1, 1);
	PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER);
#endif

#include "PlayerEngine.h"
#include "girona/RootPipeline.h"
#include "girona/Serialise.h"

static PlayerEngine* gEngine(0);
static grn::RootPipeline* gRoot(0);
static bool gExit(false);
static struct timeval last_idle_time;


#ifdef GIRONA_PSP
	#define DATA_PATH_PREFIX std::string("ms0:/")
#else
	#define DATA_PATH_PREFIX std::string("")
#endif


/* enable GLerror logging to "ms0:/log.txt" */
#ifdef GIRONA_PSP
	#define GLCHK(x)						\
	do {								\
		GLint errcode;						\
		x;							\
		errcode = glGetError();					\
		if (errcode != GL_NO_ERROR) {				\
			__psp_log("%s (%d): GL error 0x%04x\n",		\
				__FUNCTION__, __LINE__,			\
				(unsigned int) errcode);		\
		}							\
	} while (0)
#else
	#define GLCHK(x) x
#endif



static void display(void)
{
	float lDeltaTime;
	
	// Figure out time elapsed since last call to idle function
	struct timeval time_now;
	gettimeofday(&time_now, NULL);
	lDeltaTime = (float)(time_now.tv_sec  - last_idle_time.tv_sec) + 1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
	
	double lCurrentTime = 0.0;
	
	// Tell the engine the delta time
	gEngine->SetDeltaTime(lDeltaTime);
	gEngine->SetCurrentTime(lCurrentTime);

	// Run a frame
	gEngine->RunFrame();

	// Save time_now for next time
	last_idle_time = time_now;

	// Order another frame
	glutPostRedisplay();

}


static void Keyboard(unsigned char key, int , int )
{
#ifndef GIRONA_PSP	
	if (key==27) exit(0);
#else
	if (key=='x') exit(0);
#endif
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], true);
}

static void KeyboardUp(unsigned char key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutKeyMap[key], false);
}


static void SpecialKeyboard(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], true);
}

static void SpecialKeyboardUp(int key, int , int )
{
	gEngine->ProcessKey( gEngine->sGlutSpecialKeyMap[key], false);
}


static void reshape (int w, int h)
{
	grn::log("Reshaping",h);
	
	// Set up the game render pipeline
	gEngine->SetScreenArea( 0, 0, h, h );
	gEngine->Start();
}

/*
static float delta = 1.0;

static
void display (void)
{
	static GLfloat angle;
	angle += delta; 

	GLCHK(glMatrixMode(GL_MODELVIEW));
	GLCHK(glLoadIdentity());
	GLCHK(glTranslatef(0.0f, 0.0f, -2.5f));
//	GLCHK(glRotatef(angle * 0.79f, 1.0f, 0.0f, 0.0f));
//	GLCHK(glRotatef(angle * 0.98f, 0.0f, 1.0f, 0.0f));
	GLCHK(glRotatef(angle * 1.32f, 0.0f, 0.0f, 1.0f));

	GLCHK(glShadeModel(GL_SMOOTH));

	GLCHK(glClear(GL_COLOR_BUFFER_BIT));
	GLCHK(glBegin(GL_TRIANGLES));
		GLCHK(glColor3f(0.0f, 0.0f, 1.0f)); GLCHK(glVertex3f(1.0f, 0.0f, 0.0f));
		GLCHK(glColor3f(0.0f, 1.0f, 0.0f)); GLCHK(glVertex3f(0.0f, 1.0f, 0.0f));
		GLCHK(glColor3f(1.0f, 0.0f, 0.0f)); GLCHK(glVertex3f(0.0f, 0.0f, 1.0f));
	GLCHK(glEnd());
	glutSwapBuffers();
	glutPostRedisplay();
}


static
void keydown (unsigned char key, int x, int y)
{
	switch (key) {
	case 'd':			// delta, triangle 
		break;
	case 'o':			// round 
		delta = 0.0f;
		break;
	case 'q':			// square
		break;
	case 'x':			// cross button 
		exit(0);
		break;
	default:
		;
	}
}


static
void keyup (unsigned char key, int x, int y)
{
	switch (key) {
	case 'o':
		delta = 1.0f;
		break;
	default:
		;
	}
}
*/

static
void joystick (unsigned int buttonMask, int x, int y, int z)
{
	//GLCHK(glClearColor(x * 1.0f/2000.0f + 0.5f, y * 1.0f/2000.0f + 0.5f, 1.0f, 1.0f));
}


int main(int argc, char* argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode ( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow( __FILE__ );
	
	glutKeyboardFunc(Keyboard);
	glutKeyboardUpFunc(KeyboardUp);
	glutSpecialFunc (SpecialKeyboard);
	glutSpecialUpFunc (SpecialKeyboardUp);
	glutJoystickFunc(joystick, 0);
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);

	grn::log("Before initialising Girona\n");

	// Initialize the game logic
	gRoot = new grn::RootPipeline();

	grn::log("Before loading logic\n");
	
	gRoot->Load(DATA_PATH_PREFIX+"unnamed.girona");
	
	grn::log("Pipelines : ",(int)gRoot->iModules.size());

	// Create engine class
	grn::log("Before creating engine\n");
	gEngine = new PlayerEngine( gRoot );

	// Load game data
	grn::log("Befor loading data\n");

	gEngine->LoadData(DATA_PATH_PREFIX+"unnamed.data" );

	gExit = false;

	// Get the initial time
	gettimeofday (&last_idle_time, NULL);

	grn::log("After initialising Girona\n");

	glutMainLoop();

	return 0;
}

